# -*- coding: utf-8 -*-

# Copyright (c) 2021, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

import datetime
import unittest
from decimal import Decimal

from attotime import constants
from attotime.objects.attotimedelta import attotimedelta


class TestAttoTimedeltaFunctions(unittest.TestCase):
    def test_init(self):
        result = attotimedelta(weeks=1)

        self.assertEqual(result._native_timedelta, datetime.timedelta(weeks=1))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(weeks=1.5)

        self.assertEqual(result._native_timedelta, datetime.timedelta(weeks=1.5))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(weeks=-1)

        self.assertEqual(result._native_timedelta, datetime.timedelta(weeks=-1))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(weeks=-1.5)

        self.assertEqual(result._native_timedelta, datetime.timedelta(weeks=-1.5))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(days=1)

        self.assertEqual(result._native_timedelta, datetime.timedelta(days=1))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(days=1.5)

        self.assertEqual(result._native_timedelta, datetime.timedelta(days=1.5))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(days=-1)

        self.assertEqual(result._native_timedelta, datetime.timedelta(days=-1))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(days=-1.5)

        self.assertEqual(result._native_timedelta, datetime.timedelta(days=-1.5))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(hours=1)

        self.assertEqual(result._native_timedelta, datetime.timedelta(hours=1))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(hours=1.5)

        self.assertEqual(result._native_timedelta, datetime.timedelta(hours=1.5))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(hours=-1)

        self.assertEqual(result._native_timedelta, datetime.timedelta(hours=-1))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(hours=-1.5)

        self.assertEqual(result._native_timedelta, datetime.timedelta(hours=-1.5))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(minutes=1)

        self.assertEqual(result._native_timedelta, datetime.timedelta(minutes=1))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(minutes=1.5)

        self.assertEqual(result._native_timedelta, datetime.timedelta(minutes=1.5))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(minutes=-1)

        self.assertEqual(result._native_timedelta, datetime.timedelta(minutes=-1))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(minutes=-1.5)

        self.assertEqual(result._native_timedelta, datetime.timedelta(minutes=-1.5))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(seconds=1)

        self.assertEqual(result._native_timedelta, datetime.timedelta(seconds=1))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(seconds=1.5)

        self.assertEqual(result._native_timedelta, datetime.timedelta(seconds=1.5))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(seconds=-1)

        self.assertEqual(result._native_timedelta, datetime.timedelta(seconds=-1))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(seconds=-1.5)

        self.assertEqual(result._native_timedelta, datetime.timedelta(seconds=-1.5))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(milliseconds=1)

        self.assertEqual(result._native_timedelta, datetime.timedelta(milliseconds=1))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(milliseconds=1.5)

        self.assertEqual(result._native_timedelta, datetime.timedelta(milliseconds=1.5))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(milliseconds=-1)

        self.assertEqual(result._native_timedelta, datetime.timedelta(milliseconds=-1))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(milliseconds=-1.5)

        self.assertEqual(
            result._native_timedelta, datetime.timedelta(milliseconds=-1.5)
        )
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(microseconds=1)

        self.assertEqual(result._native_timedelta, datetime.timedelta(microseconds=1))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(microseconds=1.5)

        self.assertEqual(result._native_timedelta, datetime.timedelta(microseconds=1))
        self.assertEqual(result._nanoseconds, 500)

        result = attotimedelta(microseconds=-1)

        self.assertEqual(result._native_timedelta, datetime.timedelta(microseconds=-1))
        self.assertEqual(result._nanoseconds, 0)

        result = attotimedelta(microseconds=-1.5)

        self.assertEqual(result._native_timedelta, datetime.timedelta(microseconds=-1))
        self.assertEqual(result._nanoseconds, -500)

        result = attotimedelta(nanoseconds=1)

        self.assertEqual(result._native_timedelta, datetime.timedelta())
        self.assertEqual(result._nanoseconds, 1)

        result = attotimedelta(nanoseconds=1.5)

        self.assertEqual(result._native_timedelta, datetime.timedelta())
        self.assertEqual(result._nanoseconds, 1.5)

        result = attotimedelta(nanoseconds=-1)

        self.assertEqual(result._native_timedelta, datetime.timedelta())
        self.assertEqual(result._nanoseconds, -1)

        result = attotimedelta(nanoseconds=-1.5)

        self.assertEqual(result._native_timedelta, datetime.timedelta())
        self.assertEqual(result._nanoseconds, -1.5)

        result = attotimedelta(
            days=1,
            seconds=2,
            microseconds=3,
            milliseconds=4,
            minutes=5,
            hours=6,
            weeks=7,
            nanoseconds=9,
        )

        self.assertEqual(
            result._native_timedelta,
            datetime.timedelta(
                days=1,
                seconds=2,
                microseconds=3,
                milliseconds=4,
                minutes=5,
                hours=6,
                weeks=7,
            ),
        )
        self.assertEqual(result._nanoseconds, 9)

        result = attotimedelta(
            nanoseconds=1 * constants.NANOSECONDS_PER_DAY
            + 2 * constants.NANOSECONDS_PER_SECOND
            + 3 * constants.NANOSECONDS_PER_MICROSECOND
            + 4 * constants.NANOSECONDS_PER_MILLISECOND
            + 5 * constants.NANOSECONDS_PER_MINUTE
            + 6 * constants.NANOSECONDS_PER_HOUR
            + 7 * constants.NANOSECONDS_PER_WEEK
            + 9
        )

        self.assertEqual(
            result._native_timedelta,
            datetime.timedelta(
                days=1,
                seconds=2,
                microseconds=3,
                milliseconds=4,
                minutes=5,
                hours=6,
                weeks=7,
            ),
        )
        self.assertEqual(result._nanoseconds, 9)

    def test_init_type_check(self):
        with self.assertRaises(TypeError):
            attotimedelta(days="1")

        with self.assertRaises(TypeError):
            attotimedelta(days=None)

        with self.assertRaises(TypeError):
            attotimedelta(seconds="1")

        with self.assertRaises(TypeError):
            attotimedelta(seconds=None)

        with self.assertRaises(TypeError):
            attotimedelta(microseconds="1")

        with self.assertRaises(TypeError):
            attotimedelta(microseconds=None)

        with self.assertRaises(TypeError):
            attotimedelta(milliseconds="1")

        with self.assertRaises(TypeError):
            attotimedelta(milliseconds=None)

        with self.assertRaises(TypeError):
            attotimedelta(minutes="1")

        with self.assertRaises(TypeError):
            attotimedelta(minutes=None)

        with self.assertRaises(TypeError):
            attotimedelta(hours="1")

        with self.assertRaises(TypeError):
            attotimedelta(hours=None)

        with self.assertRaises(TypeError):
            attotimedelta(weeks="1")

        with self.assertRaises(TypeError):
            attotimedelta(weeks=None)

        with self.assertRaises(TypeError):
            attotimedelta(nanoseconds="1")

        with self.assertRaises(TypeError):
            attotimedelta(nanoseconds=None)

    def test_days(self):
        dt = attotimedelta(days=3)
        self.assertEqual(dt.days, 3)

        dt = attotimedelta(weeks=1)
        self.assertEqual(dt.days, constants.DAYS_PER_WEEK)

    def test_seconds(self):
        dt = attotimedelta(seconds=3)
        self.assertEqual(dt.seconds, 3)

        dt = attotimedelta(minutes=1)
        self.assertEqual(dt.seconds, constants.SECONDS_PER_MINUTE)

        dt = attotimedelta(hours=1)
        self.assertEqual(dt.seconds, constants.SECONDS_PER_HOUR)

    def test_microseconds(self):
        dt = attotimedelta(microseconds=3)
        self.assertEqual(dt.microseconds, 3)

        dt = attotimedelta(milliseconds=1)
        self.assertEqual(dt.microseconds, constants.MICROSECONDS_PER_MILLISECOND)

    def test_nanoseconds(self):
        dt = attotimedelta(nanoseconds=Decimal("3.21"))
        self.assertEqual(dt.nanoseconds, Decimal("3.21"))

    def test_add(self):
        td1 = attotimedelta(days=1)
        td2 = attotimedelta(days=2)

        result = td1 + td2
        self.assertEqual(result, attotimedelta(days=3))

        td1 = attotimedelta(days=3, nanoseconds=Decimal(4))
        td2 = attotimedelta(days=5, nanoseconds=Decimal(6))

        result = td1 + td2
        self.assertEqual(result, attotimedelta(days=8, nanoseconds=Decimal(10)))

        td1 = attotimedelta(days=5, nanoseconds=Decimal(6))
        td2 = attotimedelta(days=-3, nanoseconds=Decimal(-4))

        result = td1 + td2
        self.assertEqual(result, attotimedelta(days=2, nanoseconds=Decimal(2)))

    def test_add_overflow(self):
        td1 = attotimedelta(days=3)
        td2 = attotimedelta(days=4)

        result = td1 + td2
        self.assertEqual(result, attotimedelta(weeks=1))

        td1 = attotimedelta(seconds=30)
        td2 = attotimedelta(seconds=30)

        result = td1 + td2
        self.assertEqual(result, attotimedelta(minutes=1))

        td1 = attotimedelta(microseconds=0.5e6)
        td2 = attotimedelta(microseconds=0.5e6)

        result = td1 + td2
        self.assertEqual(result, attotimedelta(seconds=1))

        td1 = attotimedelta(milliseconds=0.5e3)
        td2 = attotimedelta(milliseconds=0.5e3)

        result = td1 + td2
        self.assertEqual(result, attotimedelta(seconds=1))

        td1 = attotimedelta(minutes=30)
        td2 = attotimedelta(minutes=30)

        result = td1 + td2
        self.assertEqual(result, attotimedelta(hours=1))

        td1 = attotimedelta(hours=12)
        td2 = attotimedelta(hours=12)

        result = td1 + td2
        self.assertEqual(result, attotimedelta(days=1))

        td1 = attotimedelta(nanoseconds=0.5e9)
        td2 = attotimedelta(nanoseconds=0.5e9)

        result = td1 + td2
        self.assertEqual(result, attotimedelta(seconds=1))

    def test_sub(self):
        td1 = attotimedelta(days=2)
        td2 = attotimedelta(days=1)

        result = td1 - td2
        self.assertEqual(result, attotimedelta(days=1))

        td1 = attotimedelta(days=5, nanoseconds=Decimal(6))
        td2 = attotimedelta(days=3, nanoseconds=Decimal(4))

        result = td1 - td2
        self.assertEqual(result, attotimedelta(days=2, nanoseconds=Decimal(2)))

        td1 = attotimedelta(days=5, nanoseconds=Decimal(6))
        td2 = attotimedelta(days=-3, nanoseconds=Decimal(-4))

        result = td1 - td2
        self.assertEqual(result, attotimedelta(days=8, nanoseconds=Decimal(10)))

    def test_sub_underflow(self):
        td1 = attotimedelta(days=1)
        td2 = attotimedelta(hours=12)

        result = td1 - td2
        self.assertEqual(result, attotimedelta(hours=12))

        td1 = attotimedelta(seconds=1)
        td2 = attotimedelta(milliseconds=0.5e3)

        result = td1 - td2
        self.assertEqual(result, attotimedelta(milliseconds=0.5e3))

        td1 = attotimedelta(microseconds=1)
        td2 = attotimedelta(nanoseconds=0.5e3)

        result = td1 - td2
        self.assertEqual(result, attotimedelta(nanoseconds=0.5e3))

        td1 = attotimedelta(milliseconds=1)
        td2 = attotimedelta(microseconds=0.5e3)

        result = td1 - td2
        self.assertEqual(result, attotimedelta(microseconds=0.5e3))

        td1 = attotimedelta(minutes=1)
        td2 = attotimedelta(seconds=30)

        result = td1 - td2
        self.assertEqual(result, attotimedelta(seconds=30))

        td1 = attotimedelta(hours=1)
        td2 = attotimedelta(minutes=30)

        result = td1 - td2
        self.assertEqual(result, attotimedelta(minutes=30))

        td1 = attotimedelta(weeks=1)
        td2 = attotimedelta(days=3)

        result = td1 - td2
        self.assertEqual(result, attotimedelta(days=4))

        td1 = attotimedelta(nanoseconds=1)
        td2 = attotimedelta(nanoseconds=0.5)

        result = td1 - td2
        self.assertEqual(result, attotimedelta(nanoseconds=0.5))

    def test_mul(self):
        result = attotimedelta(days=2) * 3
        self.assertEqual(result, attotimedelta(days=6))

        result = attotimedelta(days=5, nanoseconds=Decimal(6)) * 3
        self.assertEqual(result, attotimedelta(days=15, nanoseconds=Decimal(18)))

        result = attotimedelta(days=5, nanoseconds=Decimal(6)) * -3
        self.assertEqual(result, attotimedelta(days=-15, nanoseconds=Decimal(-18)))

        result = attotimedelta(days=2) * 0.5
        self.assertEqual(result, attotimedelta(days=1))

        result = attotimedelta(days=4) * Decimal(0.25)
        self.assertEqual(result, attotimedelta(days=1))

        result = 3 * attotimedelta(days=2)
        self.assertEqual(result, attotimedelta(days=6))

        result = 3 * attotimedelta(days=5, nanoseconds=Decimal(6))
        self.assertEqual(result, attotimedelta(days=15, nanoseconds=Decimal(18)))

        result = -3 * attotimedelta(days=5, nanoseconds=Decimal(6))
        self.assertEqual(result, attotimedelta(days=-15, nanoseconds=Decimal(-18)))

        result = 0.5 * attotimedelta(days=2)
        self.assertEqual(result, attotimedelta(days=1))

        result = Decimal(0.25) * attotimedelta(days=4)
        self.assertEqual(result, attotimedelta(days=1))

    def test_mul_overflow(self):
        result = attotimedelta(days=1) * 7
        self.assertEqual(result, attotimedelta(weeks=1))

        result = attotimedelta(seconds=30) * 2
        self.assertEqual(result, attotimedelta(minutes=1))

        result = attotimedelta(microseconds=500) * 2
        self.assertEqual(result, attotimedelta(milliseconds=1))

        result = attotimedelta(milliseconds=500) * 2
        self.assertEqual(result, attotimedelta(seconds=1))

        result = attotimedelta(minutes=30) * 2
        self.assertEqual(result, attotimedelta(hours=1))

        result = attotimedelta(hours=12) * 2
        self.assertEqual(result, attotimedelta(days=1))

        result = attotimedelta(nanoseconds=500) * 2
        self.assertEqual(result, attotimedelta(microseconds=1))

    def test_mul_underflow(self):
        result = attotimedelta(days=1) * 0.5
        self.assertEqual(result, attotimedelta(hours=12))

        result = attotimedelta(seconds=1) * Decimal("0.1")
        self.assertEqual(result, attotimedelta(milliseconds=100))

        result = attotimedelta(microseconds=1) * Decimal("0.1")
        self.assertEqual(result, attotimedelta(nanoseconds=100))

        result = attotimedelta(milliseconds=1) * Decimal("0.1")
        self.assertEqual(result, attotimedelta(microseconds=100))

        result = attotimedelta(minutes=1) * 0.5
        self.assertEqual(result, attotimedelta(seconds=30))

        result = attotimedelta(hours=1) * 0.5
        self.assertEqual(result, attotimedelta(minutes=30))

        result = attotimedelta(nanoseconds=1) * 0.5
        self.assertEqual(result, attotimedelta(nanoseconds=0.5))

    def test_div(self):
        result = attotimedelta(days=2) / 2
        self.assertEqual(result, attotimedelta(days=1))

        result = attotimedelta(days=6, nanoseconds=Decimal(6)) / 3
        self.assertEqual(result, attotimedelta(days=2, nanoseconds=Decimal(2)))

        result = attotimedelta(days=6, nanoseconds=Decimal(6)) / -2
        self.assertEqual(result, attotimedelta(days=-3, nanoseconds=Decimal(-3)))

        result = attotimedelta(days=2) / 0.5
        self.assertEqual(result, attotimedelta(days=4))

        result = attotimedelta(days=4) / Decimal(0.25)
        self.assertEqual(result, attotimedelta(days=16))

        result = attotimedelta(nanoseconds=5) / 2
        self.assertEqual(result, attotimedelta(nanoseconds=Decimal(2.5)))

    def test_div_overflow(self):
        result = attotimedelta(days=4) / (Decimal(2) / Decimal(7))
        self.assertEqual(result, attotimedelta(weeks=2))

        result = attotimedelta(seconds=30) / Decimal(0.5)
        self.assertEqual(result, attotimedelta(minutes=1))

        result = attotimedelta(microseconds=500) / Decimal(0.5)
        self.assertEqual(result, attotimedelta(milliseconds=1))

        result = attotimedelta(milliseconds=500) / Decimal(0.5)
        self.assertEqual(result, attotimedelta(seconds=1))

        result = attotimedelta(minutes=30) / Decimal(0.5)
        self.assertEqual(result, attotimedelta(hours=1))

        result = attotimedelta(hours=12) / Decimal(0.5)
        self.assertEqual(result, attotimedelta(days=1))

        result = attotimedelta(nanoseconds=500) / Decimal(0.5)
        self.assertEqual(result, attotimedelta(microseconds=1))

    def test_div_underflow(self):
        result = attotimedelta(days=1) / 2
        self.assertEqual(result, attotimedelta(hours=12))

        result = attotimedelta(seconds=1) / 10
        self.assertEqual(result, attotimedelta(milliseconds=100))

        result = attotimedelta(microseconds=1) / 10
        self.assertEqual(result, attotimedelta(nanoseconds=100))

        result = attotimedelta(milliseconds=1) / 10
        self.assertEqual(result, attotimedelta(microseconds=100))

        result = attotimedelta(minutes=1) / 2
        self.assertEqual(result, attotimedelta(seconds=30))

        result = attotimedelta(hours=1) / 2
        self.assertEqual(result, attotimedelta(minutes=30))

    def test_floordiv(self):
        result = attotimedelta(days=2) // 2
        self.assertEqual(result, attotimedelta(days=1))

        result = attotimedelta(days=6, nanoseconds=Decimal(6)) // 3
        self.assertEqual(result, attotimedelta(days=2, nanoseconds=Decimal(2)))

        result = attotimedelta(days=6, nanoseconds=Decimal(6)) // -2
        self.assertEqual(result, attotimedelta(days=-3, nanoseconds=Decimal(-3)))

        result = attotimedelta(days=2) // 0.5
        self.assertEqual(result, attotimedelta(days=4))

        result = attotimedelta(days=4) // Decimal(0.25)
        self.assertEqual(result, attotimedelta(days=16))

        result = attotimedelta(nanoseconds=5) // 2
        self.assertEqual(result, attotimedelta(nanoseconds=2))

    def test_floordiv_overflow(self):
        result = attotimedelta(days=4) // (Decimal(2) / Decimal(7))
        self.assertEqual(result, attotimedelta(weeks=2))

        result = attotimedelta(seconds=30) // Decimal(0.5)
        self.assertEqual(result, attotimedelta(minutes=1))

        result = attotimedelta(microseconds=500) // Decimal(0.5)
        self.assertEqual(result, attotimedelta(milliseconds=1))

        result = attotimedelta(milliseconds=500) // Decimal(0.5)
        self.assertEqual(result, attotimedelta(seconds=1))

        result = attotimedelta(minutes=30) // Decimal(0.5)
        self.assertEqual(result, attotimedelta(hours=1))

        result = attotimedelta(hours=12) // Decimal(0.5)
        self.assertEqual(result, attotimedelta(days=1))

        result = attotimedelta(nanoseconds=500) // Decimal(0.5)
        self.assertEqual(result, attotimedelta(microseconds=1))

    def test_floordiv_underflow(self):
        result = attotimedelta(days=1) // 2
        self.assertEqual(result, attotimedelta(hours=12))

        result = attotimedelta(seconds=1) // 10
        self.assertEqual(result, attotimedelta(milliseconds=100))

        result = attotimedelta(microseconds=1) // 10
        self.assertEqual(result, attotimedelta(nanoseconds=100))

        result = attotimedelta(milliseconds=1) // 10
        self.assertEqual(result, attotimedelta(microseconds=100))

        result = attotimedelta(minutes=1) // 2
        self.assertEqual(result, attotimedelta(seconds=30))

        result = attotimedelta(hours=1) // 2
        self.assertEqual(result, attotimedelta(minutes=30))

    def test_eq(self):
        self.assertTrue(attotimedelta(days=1) == attotimedelta(days=1))
        self.assertTrue(attotimedelta(seconds=2) == attotimedelta(seconds=2))
        self.assertTrue(attotimedelta(microseconds=3) == attotimedelta(microseconds=3))
        self.assertTrue(attotimedelta(milliseconds=4) == attotimedelta(milliseconds=4))
        self.assertTrue(attotimedelta(minutes=5) == attotimedelta(minutes=5))
        self.assertTrue(attotimedelta(hours=6) == attotimedelta(hours=6))
        self.assertTrue(attotimedelta(weeks=7) == attotimedelta(weeks=7))
        self.assertTrue(
            attotimedelta(nanoseconds=Decimal(8))
            == attotimedelta(nanoseconds=Decimal(8))
        )
        self.assertTrue(
            attotimedelta(
                days=1,
                seconds=2,
                microseconds=3,
                milliseconds=4,
                minutes=5,
                hours=6,
                weeks=7,
                nanoseconds=Decimal(8),
            )
            == attotimedelta(
                days=1,
                seconds=2,
                microseconds=3,
                milliseconds=4,
                minutes=5,
                hours=6,
                weeks=7,
                nanoseconds=Decimal(8),
            )
        )

        # Test some equivalences
        # TODO: 1/7 of a week is not equal to 1 day due to rounding errors
        self.assertTrue(attotimedelta(days=7) == attotimedelta(weeks=1))
        self.assertTrue(attotimedelta(seconds=60) == attotimedelta(minutes=1))
        self.assertTrue(attotimedelta(microseconds=1 * 1e6) == attotimedelta(seconds=1))
        self.assertTrue(attotimedelta(milliseconds=1 * 1e3) == attotimedelta(seconds=1))
        self.assertTrue(attotimedelta(minutes=60) == attotimedelta(hours=1))
        self.assertTrue(attotimedelta(nanoseconds=1 * 1e9) == attotimedelta(seconds=1))

        # Not equals
        self.assertFalse(attotimedelta(days=1) == attotimedelta(days=2))
        self.assertFalse(attotimedelta(seconds=3) == attotimedelta(seconds=4))
        self.assertFalse(attotimedelta(microseconds=5) == attotimedelta(microseconds=6))
        self.assertFalse(attotimedelta(milliseconds=7) == attotimedelta(milliseconds=8))
        self.assertFalse(attotimedelta(minutes=9) == attotimedelta(minutes=10))
        self.assertFalse(attotimedelta(hours=11) == attotimedelta(hours=12))
        self.assertFalse(attotimedelta(weeks=13) == attotimedelta(weeks=14))
        self.assertFalse(
            attotimedelta(nanoseconds=Decimal(15))
            == attotimedelta(nanoseconds=Decimal(16))
        )
        self.assertFalse(
            attotimedelta(
                days=1,
                seconds=2,
                microseconds=3,
                milliseconds=4,
                minutes=5,
                hours=6,
                weeks=7,
                nanoseconds=Decimal(8),
            )
            == attotimedelta(
                days=9,
                seconds=10,
                microseconds=11,
                milliseconds=12,
                minutes=13,
                hours=14,
                weeks=15,
                nanoseconds=Decimal(16),
            )
        )

        # Class check
        self.assertFalse(attotimedelta(days=1) == datetime.timedelta(days=1))

    def test_ne(self):
        # Inverse of test_eq
        self.assertFalse(attotimedelta(days=1) != attotimedelta(days=1))
        self.assertFalse(attotimedelta(seconds=2) != attotimedelta(seconds=2))
        self.assertFalse(attotimedelta(microseconds=3) != attotimedelta(microseconds=3))
        self.assertFalse(attotimedelta(milliseconds=4) != attotimedelta(milliseconds=4))
        self.assertFalse(attotimedelta(minutes=5) != attotimedelta(minutes=5))
        self.assertFalse(attotimedelta(hours=6) != attotimedelta(hours=6))
        self.assertFalse(attotimedelta(weeks=7) != attotimedelta(weeks=7))
        self.assertFalse(
            attotimedelta(nanoseconds=Decimal(8))
            != attotimedelta(nanoseconds=Decimal(8))
        )
        self.assertFalse(
            attotimedelta(
                days=1,
                seconds=2,
                microseconds=3,
                milliseconds=4,
                minutes=5,
                hours=6,
                weeks=7,
                nanoseconds=Decimal(8),
            )
            != attotimedelta(
                days=1,
                seconds=2,
                microseconds=3,
                milliseconds=4,
                minutes=5,
                hours=6,
                weeks=7,
                nanoseconds=Decimal(8),
            )
        )

        # Test some equivalences
        # TODO: 1/7 of a week is not equal to 1 day due to rounding errors
        self.assertFalse(attotimedelta(days=7) != attotimedelta(weeks=1))
        self.assertFalse(attotimedelta(seconds=60) != attotimedelta(minutes=1))
        self.assertFalse(
            attotimedelta(microseconds=1 * 1e6) != attotimedelta(seconds=1)
        )
        self.assertFalse(
            attotimedelta(milliseconds=1 * 1e3) != attotimedelta(seconds=1)
        )
        self.assertFalse(attotimedelta(minutes=60) != attotimedelta(hours=1))
        self.assertFalse(attotimedelta(nanoseconds=1 * 1e9) != attotimedelta(seconds=1))

        # Not equals
        self.assertTrue(attotimedelta(days=1) != attotimedelta(days=2))
        self.assertTrue(attotimedelta(seconds=3) != attotimedelta(seconds=4))
        self.assertTrue(attotimedelta(microseconds=5) != attotimedelta(microseconds=6))
        self.assertTrue(attotimedelta(milliseconds=7) != attotimedelta(milliseconds=8))
        self.assertTrue(attotimedelta(minutes=9) != attotimedelta(minutes=10))
        self.assertTrue(attotimedelta(hours=11) != attotimedelta(hours=12))
        self.assertTrue(attotimedelta(weeks=13) != attotimedelta(weeks=14))
        self.assertTrue(
            attotimedelta(nanoseconds=Decimal(15))
            != attotimedelta(nanoseconds=Decimal(16))
        )
        self.assertTrue(
            attotimedelta(
                days=1,
                seconds=2,
                microseconds=3,
                milliseconds=4,
                minutes=5,
                hours=6,
                weeks=7,
                nanoseconds=Decimal(8),
            )
            != attotimedelta(
                days=9,
                seconds=10,
                microseconds=11,
                milliseconds=12,
                minutes=13,
                hours=14,
                weeks=15,
                nanoseconds=Decimal(16),
            )
        )

        # Class check
        self.assertTrue(attotimedelta(days=1) != datetime.timedelta(days=1))

    def test_gt(self):
        self.assertTrue(attotimedelta(days=2) > attotimedelta(days=1))
        self.assertTrue(attotimedelta(seconds=2) > attotimedelta(seconds=1))
        self.assertTrue(attotimedelta(microseconds=2) > attotimedelta(microseconds=1))
        self.assertTrue(attotimedelta(milliseconds=2) > attotimedelta(milliseconds=1))
        self.assertTrue(attotimedelta(minutes=2) > attotimedelta(minutes=1))
        self.assertTrue(attotimedelta(hours=2) > attotimedelta(hours=1))
        self.assertTrue(attotimedelta(weeks=2) > attotimedelta(weeks=1))
        self.assertTrue(
            attotimedelta(nanoseconds=Decimal(2))
            > attotimedelta(nanoseconds=Decimal(1))
        )
        self.assertTrue(
            attotimedelta(
                days=2,
                seconds=2,
                microseconds=2,
                milliseconds=2,
                minutes=2,
                hours=2,
                weeks=2,
                nanoseconds=Decimal(2),
            )
            > attotimedelta(
                days=1,
                seconds=1,
                microseconds=1,
                milliseconds=1,
                minutes=1,
                hours=1,
                weeks=1,
                nanoseconds=Decimal(1),
            )
        )

        # Test some equivalences
        self.assertFalse(attotimedelta(days=2) > attotimedelta(days=2))
        self.assertFalse(attotimedelta(seconds=2) > attotimedelta(seconds=2))
        self.assertFalse(attotimedelta(microseconds=2) > attotimedelta(microseconds=2))
        self.assertFalse(attotimedelta(milliseconds=2) > attotimedelta(milliseconds=2))
        self.assertFalse(attotimedelta(minutes=2) > attotimedelta(minutes=2))
        self.assertFalse(attotimedelta(hours=2) > attotimedelta(hours=2))
        self.assertFalse(attotimedelta(weeks=2) > attotimedelta(weeks=2))
        self.assertFalse(
            attotimedelta(nanoseconds=Decimal(2))
            > attotimedelta(nanoseconds=Decimal(2))
        )
        self.assertFalse(
            attotimedelta(
                days=2,
                seconds=2,
                microseconds=2,
                milliseconds=2,
                minutes=2,
                hours=2,
                weeks=2,
                nanoseconds=Decimal(2),
            )
            > attotimedelta(
                days=2,
                seconds=2,
                microseconds=2,
                milliseconds=2,
                minutes=2,
                hours=2,
                weeks=2,
                nanoseconds=Decimal(2),
            )
        )

        # Not greaters
        self.assertFalse(attotimedelta(days=1) > attotimedelta(days=2))
        self.assertFalse(attotimedelta(seconds=1) > attotimedelta(seconds=4))
        self.assertFalse(attotimedelta(microseconds=1) > attotimedelta(microseconds=6))
        self.assertFalse(attotimedelta(milliseconds=1) > attotimedelta(milliseconds=8))
        self.assertFalse(attotimedelta(minutes=1) > attotimedelta(minutes=10))
        self.assertFalse(attotimedelta(hours=1) > attotimedelta(hours=12))
        self.assertFalse(attotimedelta(weeks=1) > attotimedelta(weeks=14))
        self.assertFalse(
            attotimedelta(nanoseconds=Decimal(1))
            > attotimedelta(nanoseconds=Decimal(16))
        )
        self.assertFalse(
            attotimedelta(
                days=1,
                seconds=1,
                microseconds=1,
                milliseconds=1,
                minutes=1,
                hours=1,
                weeks=1,
                nanoseconds=Decimal(1),
            )
            > attotimedelta(
                days=2,
                seconds=2,
                microseconds=2,
                milliseconds=2,
                minutes=2,
                hours=2,
                weeks=2,
                nanoseconds=Decimal(2),
            )
        )

        # Class check
        with self.assertRaises(TypeError):
            attotimedelta(days=1) > 1

    def test_ge(self):
        self.assertTrue(attotimedelta(days=2) >= attotimedelta(days=1))
        self.assertTrue(attotimedelta(seconds=2) >= attotimedelta(seconds=1))
        self.assertTrue(attotimedelta(microseconds=2) >= attotimedelta(microseconds=1))
        self.assertTrue(attotimedelta(milliseconds=2) >= attotimedelta(milliseconds=1))
        self.assertTrue(attotimedelta(minutes=2) >= attotimedelta(minutes=1))
        self.assertTrue(attotimedelta(hours=2) >= attotimedelta(hours=1))
        self.assertTrue(attotimedelta(weeks=2) >= attotimedelta(weeks=1))
        self.assertTrue(
            attotimedelta(nanoseconds=Decimal(2))
            >= attotimedelta(nanoseconds=Decimal(1))
        )
        self.assertTrue(
            attotimedelta(
                days=2,
                seconds=2,
                microseconds=2,
                milliseconds=2,
                minutes=2,
                hours=2,
                weeks=2,
                nanoseconds=Decimal(2),
            )
            >= attotimedelta(
                days=1,
                seconds=1,
                microseconds=1,
                milliseconds=1,
                minutes=1,
                hours=1,
                weeks=1,
                nanoseconds=Decimal(1),
            )
        )

        # Test some equivalences
        self.assertTrue(attotimedelta(days=2) >= attotimedelta(days=2))
        self.assertTrue(attotimedelta(seconds=2) >= attotimedelta(seconds=2))
        self.assertTrue(attotimedelta(microseconds=2) >= attotimedelta(microseconds=2))
        self.assertTrue(attotimedelta(milliseconds=2) >= attotimedelta(milliseconds=2))
        self.assertTrue(attotimedelta(minutes=2) >= attotimedelta(minutes=2))
        self.assertTrue(attotimedelta(hours=2) >= attotimedelta(hours=2))
        self.assertTrue(attotimedelta(weeks=2) >= attotimedelta(weeks=2))
        self.assertTrue(
            attotimedelta(nanoseconds=Decimal(2))
            >= attotimedelta(nanoseconds=Decimal(2))
        )
        self.assertTrue(
            attotimedelta(
                days=2,
                seconds=2,
                microseconds=2,
                milliseconds=2,
                minutes=2,
                hours=2,
                weeks=2,
                nanoseconds=Decimal(2),
            )
            >= attotimedelta(
                days=2,
                seconds=2,
                microseconds=2,
                milliseconds=2,
                minutes=2,
                hours=2,
                weeks=2,
                nanoseconds=Decimal(2),
            )
        )

        # Not greaters
        self.assertFalse(attotimedelta(days=1) >= attotimedelta(days=2))
        self.assertFalse(attotimedelta(seconds=1) >= attotimedelta(seconds=4))
        self.assertFalse(attotimedelta(microseconds=1) >= attotimedelta(microseconds=6))
        self.assertFalse(attotimedelta(milliseconds=1) >= attotimedelta(milliseconds=8))
        self.assertFalse(attotimedelta(minutes=1) >= attotimedelta(minutes=10))
        self.assertFalse(attotimedelta(hours=1) >= attotimedelta(hours=12))
        self.assertFalse(attotimedelta(weeks=1) >= attotimedelta(weeks=14))
        self.assertFalse(
            attotimedelta(nanoseconds=Decimal(1))
            >= attotimedelta(nanoseconds=Decimal(16))
        )
        self.assertFalse(
            attotimedelta(
                days=1,
                seconds=1,
                microseconds=1,
                milliseconds=1,
                minutes=1,
                hours=1,
                weeks=1,
                nanoseconds=Decimal(1),
            )
            >= attotimedelta(
                days=2,
                seconds=2,
                microseconds=2,
                milliseconds=2,
                minutes=2,
                hours=2,
                weeks=2,
                nanoseconds=Decimal(2),
            )
        )

        # Class check
        with self.assertRaises(TypeError):
            attotimedelta(days=1) >= 1

    def test_lt(self):
        self.assertTrue(attotimedelta(days=1) < attotimedelta(days=2))
        self.assertTrue(attotimedelta(seconds=1) < attotimedelta(seconds=2))
        self.assertTrue(attotimedelta(microseconds=1) < attotimedelta(microseconds=2))
        self.assertTrue(attotimedelta(milliseconds=1) < attotimedelta(milliseconds=2))
        self.assertTrue(attotimedelta(minutes=1) < attotimedelta(minutes=2))
        self.assertTrue(attotimedelta(hours=1) < attotimedelta(hours=2))
        self.assertTrue(attotimedelta(weeks=1) < attotimedelta(weeks=2))
        self.assertTrue(
            attotimedelta(nanoseconds=Decimal(1))
            < attotimedelta(nanoseconds=Decimal(2))
        )
        self.assertTrue(
            attotimedelta(
                days=1,
                seconds=1,
                microseconds=1,
                milliseconds=1,
                minutes=1,
                hours=1,
                weeks=1,
                nanoseconds=Decimal(1),
            )
            < attotimedelta(
                days=2,
                seconds=2,
                microseconds=2,
                milliseconds=2,
                minutes=2,
                hours=2,
                weeks=2,
                nanoseconds=Decimal(2),
            )
        )

        # Test some equivalences
        self.assertFalse(attotimedelta(days=2) < attotimedelta(days=2))
        self.assertFalse(attotimedelta(seconds=2) < attotimedelta(seconds=2))
        self.assertFalse(attotimedelta(microseconds=2) < attotimedelta(microseconds=2))
        self.assertFalse(attotimedelta(milliseconds=2) < attotimedelta(milliseconds=2))
        self.assertFalse(attotimedelta(minutes=2) < attotimedelta(minutes=2))
        self.assertFalse(attotimedelta(hours=2) < attotimedelta(hours=2))
        self.assertFalse(attotimedelta(weeks=2) < attotimedelta(weeks=2))
        self.assertFalse(
            attotimedelta(nanoseconds=Decimal(2))
            < attotimedelta(nanoseconds=Decimal(2))
        )
        self.assertFalse(
            attotimedelta(
                days=2,
                seconds=2,
                microseconds=2,
                milliseconds=2,
                minutes=2,
                hours=2,
                weeks=2,
                nanoseconds=Decimal(2),
            )
            < attotimedelta(
                days=2,
                seconds=2,
                microseconds=2,
                milliseconds=2,
                minutes=2,
                hours=2,
                weeks=2,
                nanoseconds=Decimal(2),
            )
        )

        # Not greaters
        self.assertFalse(attotimedelta(days=2) < attotimedelta(days=1))
        self.assertFalse(attotimedelta(seconds=4) < attotimedelta(seconds=1))
        self.assertFalse(attotimedelta(microseconds=6) < attotimedelta(microseconds=1))
        self.assertFalse(attotimedelta(milliseconds=8) < attotimedelta(milliseconds=1))
        self.assertFalse(attotimedelta(minutes=10) < attotimedelta(minutes=1))
        self.assertFalse(attotimedelta(hours=12) < attotimedelta(hours=1))
        self.assertFalse(attotimedelta(weeks=14) < attotimedelta(weeks=1))
        self.assertFalse(
            attotimedelta(nanoseconds=Decimal(16))
            < attotimedelta(nanoseconds=Decimal(1))
        )
        self.assertFalse(
            attotimedelta(
                days=2,
                seconds=2,
                microseconds=2,
                milliseconds=2,
                minutes=2,
                hours=2,
                weeks=2,
                nanoseconds=Decimal(2),
            )
            < attotimedelta(
                days=1,
                seconds=1,
                microseconds=1,
                milliseconds=1,
                minutes=1,
                hours=1,
                weeks=1,
                nanoseconds=Decimal(1),
            )
        )

        # Class check
        with self.assertRaises(TypeError):
            attotimedelta(days=1) < 1

    def test_le(self):
        self.assertTrue(attotimedelta(days=1) <= attotimedelta(days=2))
        self.assertTrue(attotimedelta(seconds=1) <= attotimedelta(seconds=2))
        self.assertTrue(attotimedelta(microseconds=1) <= attotimedelta(microseconds=2))
        self.assertTrue(attotimedelta(milliseconds=1) <= attotimedelta(milliseconds=2))
        self.assertTrue(attotimedelta(minutes=1) <= attotimedelta(minutes=2))
        self.assertTrue(attotimedelta(hours=1) <= attotimedelta(hours=2))
        self.assertTrue(attotimedelta(weeks=1) <= attotimedelta(weeks=2))
        self.assertTrue(
            attotimedelta(nanoseconds=Decimal(1))
            <= attotimedelta(nanoseconds=Decimal(2))
        )
        self.assertTrue(
            attotimedelta(
                days=1,
                seconds=1,
                microseconds=1,
                milliseconds=1,
                minutes=1,
                hours=1,
                weeks=1,
                nanoseconds=Decimal(1),
            )
            <= attotimedelta(
                days=2,
                seconds=2,
                microseconds=2,
                milliseconds=2,
                minutes=2,
                hours=2,
                weeks=2,
                nanoseconds=Decimal(2),
            )
        )

        # Test some equivalences
        self.assertTrue(attotimedelta(days=2) <= attotimedelta(days=2))
        self.assertTrue(attotimedelta(seconds=2) <= attotimedelta(seconds=2))
        self.assertTrue(attotimedelta(microseconds=2) <= attotimedelta(microseconds=2))
        self.assertTrue(attotimedelta(milliseconds=2) <= attotimedelta(milliseconds=2))
        self.assertTrue(attotimedelta(minutes=2) <= attotimedelta(minutes=2))
        self.assertTrue(attotimedelta(hours=2) <= attotimedelta(hours=2))
        self.assertTrue(attotimedelta(weeks=2) <= attotimedelta(weeks=2))
        self.assertTrue(
            attotimedelta(nanoseconds=Decimal(2))
            <= attotimedelta(nanoseconds=Decimal(2))
        )
        self.assertTrue(
            attotimedelta(
                days=2,
                seconds=2,
                microseconds=2,
                milliseconds=2,
                minutes=2,
                hours=2,
                weeks=2,
                nanoseconds=Decimal(2),
            )
            <= attotimedelta(
                days=2,
                seconds=2,
                microseconds=2,
                milliseconds=2,
                minutes=2,
                hours=2,
                weeks=2,
                nanoseconds=Decimal(2),
            )
        )

        # Not greaters
        self.assertFalse(attotimedelta(days=2) <= attotimedelta(days=1))
        self.assertFalse(attotimedelta(seconds=4) <= attotimedelta(seconds=1))
        self.assertFalse(attotimedelta(microseconds=6) <= attotimedelta(microseconds=1))
        self.assertFalse(attotimedelta(milliseconds=8) <= attotimedelta(milliseconds=1))
        self.assertFalse(attotimedelta(minutes=10) <= attotimedelta(minutes=1))
        self.assertFalse(attotimedelta(hours=12) <= attotimedelta(hours=1))
        self.assertFalse(attotimedelta(weeks=14) <= attotimedelta(weeks=1))
        self.assertFalse(
            attotimedelta(nanoseconds=Decimal(16))
            <= attotimedelta(nanoseconds=Decimal(1))
        )
        self.assertFalse(
            attotimedelta(
                days=2,
                seconds=2,
                microseconds=2,
                milliseconds=2,
                minutes=2,
                hours=2,
                weeks=2,
                nanoseconds=Decimal(2),
            )
            <= attotimedelta(
                days=1,
                seconds=1,
                microseconds=1,
                milliseconds=1,
                minutes=1,
                hours=1,
                weeks=1,
                nanoseconds=Decimal(1),
            )
        )

        # Class check
        with self.assertRaises(TypeError):
            attotimedelta(days=1) <= 1

    def test_pos(self):
        td1 = attotimedelta(days=5, nanoseconds=Decimal(6))
        td2 = attotimedelta(days=-3, nanoseconds=Decimal(-4))
        # https://stackoverflow.com/questions/16819023/whats-the-purpose-of-the-pos-unary-operator-in-python
        td3 = attotimedelta(
            nanoseconds=Decimal("3.1415926535897932384626433832795028841971")
        )

        result = +td1
        self.assertEqual(result, td1)

        result = +td2
        self.assertEqual(result, td2)

        result = +td3
        self.assertEqual(
            result, attotimedelta(nanoseconds=Decimal("3.141592653589793238462643383"))
        )

    def test_neg(self):
        td1 = attotimedelta(days=5, nanoseconds=Decimal(6))
        td2 = attotimedelta(days=-3, nanoseconds=Decimal(-4))
        # https://stackoverflow.com/questions/16819023/whats-the-purpose-of-the-pos-unary-operator-in-python
        td3 = attotimedelta(
            nanoseconds=Decimal("3.1415926535897932384626433832795028841971")
        )

        result = -td1
        self.assertEqual(result, attotimedelta(days=-5, nanoseconds=Decimal(-6)))

        result = -td2
        self.assertEqual(result, attotimedelta(days=3, nanoseconds=Decimal(4)))

        result = -td3
        self.assertEqual(
            result, attotimedelta(nanoseconds=Decimal("-3.141592653589793238462643383"))
        )

    def test_abs(self):
        td1 = attotimedelta(days=5, nanoseconds=Decimal(6))
        td2 = attotimedelta(days=-3, nanoseconds=Decimal(-4))
        td3 = attotimedelta(days=-3, nanoseconds=Decimal(4))
        td4 = attotimedelta(days=3, nanoseconds=Decimal(-4))

        result = abs(td1)
        self.assertEqual(result, attotimedelta(days=5, nanoseconds=Decimal(6)))

        result = abs(td2)
        self.assertEqual(result, attotimedelta(days=3, nanoseconds=Decimal(4)))

        result = abs(td3)
        self.assertEqual(
            result, attotimedelta(days=3) - attotimedelta(nanoseconds=Decimal(4))
        )

        result = abs(td4)
        self.assertEqual(
            result, attotimedelta(days=3) - attotimedelta(nanoseconds=Decimal(4))
        )

    def test_str(self):
        result = str(attotimedelta(days=1))
        self.assertEqual(result, "1 day, 0:00:00")

        result = str(attotimedelta(days=5))
        self.assertEqual(result, "5 days, 0:00:00")

        result = str(attotimedelta(days=50))
        self.assertEqual(result, "50 days, 0:00:00")

        result = str(attotimedelta(days=-1))
        self.assertEqual(result, "-1 day, 0:00:00")

        result = str(attotimedelta(days=-5))
        self.assertEqual(result, "-5 days, 0:00:00")

        result = str(attotimedelta(days=-50))
        self.assertEqual(result, "-50 days, 0:00:00")

        result = str(attotimedelta(seconds=1))
        self.assertEqual(result, "0:00:01")

        result = str(attotimedelta(seconds=5))
        self.assertEqual(result, "0:00:05")

        result = str(attotimedelta(seconds=50))
        self.assertEqual(result, "0:00:50")

        result = str(attotimedelta(seconds=-1))
        self.assertEqual(result, "-1 day, 23:59:59")

        result = str(attotimedelta(seconds=-5))
        self.assertEqual(result, "-1 day, 23:59:55")

        result = str(attotimedelta(seconds=-50))
        self.assertEqual(result, "-1 day, 23:59:10")

        result = str(attotimedelta(microseconds=1))
        self.assertEqual(result, "0:00:00.000001")

        result = str(attotimedelta(microseconds=5))
        self.assertEqual(result, "0:00:00.000005")

        result = str(attotimedelta(microseconds=50))
        self.assertEqual(result, "0:00:00.00005")

        result = str(attotimedelta(microseconds=-1))
        self.assertEqual(result, "-1 day, 23:59:59.999999")

        result = str(attotimedelta(microseconds=-5))
        self.assertEqual(result, "-1 day, 23:59:59.999995")

        result = str(attotimedelta(microseconds=-50))
        self.assertEqual(result, "-1 day, 23:59:59.99995")

        result = str(attotimedelta(milliseconds=1))
        self.assertEqual(result, "0:00:00.001")

        result = str(attotimedelta(milliseconds=5))
        self.assertEqual(result, "0:00:00.005")

        result = str(attotimedelta(milliseconds=50))
        self.assertEqual(result, "0:00:00.05")

        result = str(attotimedelta(milliseconds=-1))
        self.assertEqual(result, "-1 day, 23:59:59.999")

        result = str(attotimedelta(milliseconds=-5))
        self.assertEqual(result, "-1 day, 23:59:59.995")

        result = str(attotimedelta(milliseconds=-50))
        self.assertEqual(result, "-1 day, 23:59:59.95")

        result = str(attotimedelta(minutes=1))
        self.assertEqual(result, "0:01:00")

        result = str(attotimedelta(minutes=5))
        self.assertEqual(result, "0:05:00")

        result = str(attotimedelta(minutes=50))
        self.assertEqual(result, "0:50:00")

        result = str(attotimedelta(minutes=-1))
        self.assertEqual(result, "-1 day, 23:59:00")

        result = str(attotimedelta(minutes=-5))
        self.assertEqual(result, "-1 day, 23:55:00")

        result = str(attotimedelta(minutes=-50))
        self.assertEqual(result, "-1 day, 23:10:00")

        result = str(attotimedelta(hours=1))
        self.assertEqual(result, "1:00:00")

        result = str(attotimedelta(hours=5))
        self.assertEqual(result, "5:00:00")

        result = str(attotimedelta(hours=12))
        self.assertEqual(result, "12:00:00")

        result = str(attotimedelta(hours=-1))
        self.assertEqual(result, "-1 day, 23:00:00")

        result = str(attotimedelta(hours=-5))
        self.assertEqual(result, "-1 day, 19:00:00")

        result = str(attotimedelta(hours=-12))
        self.assertEqual(result, "-1 day, 12:00:00")

        result = str(attotimedelta(weeks=1))
        self.assertEqual(result, "7 days, 0:00:00")

        result = str(attotimedelta(weeks=5))
        self.assertEqual(result, "35 days, 0:00:00")

        result = str(attotimedelta(weeks=50))
        self.assertEqual(result, "350 days, 0:00:00")

        result = str(attotimedelta(weeks=-1))
        self.assertEqual(result, "-7 days, 0:00:00")

        result = str(attotimedelta(weeks=-5))
        self.assertEqual(result, "-35 days, 0:00:00")

        result = str(attotimedelta(weeks=-50))
        self.assertEqual(result, "-350 days, 0:00:00")

        result = str(attotimedelta(nanoseconds=1))
        self.assertEqual(result, "0:00:00.000000001")

        result = str(attotimedelta(nanoseconds=5))
        self.assertEqual(result, "0:00:00.000000005")

        result = str(attotimedelta(nanoseconds=50))
        self.assertEqual(result, "0:00:00.00000005")

        result = str(attotimedelta(nanoseconds=Decimal("0.001")))
        self.assertEqual(result, "0:00:00.000000000001")

        result = str(attotimedelta(nanoseconds=-1))
        self.assertEqual(result, "-1 day, 23:59:59.999999999")

        result = str(attotimedelta(nanoseconds=-5))
        self.assertEqual(result, "-1 day, 23:59:59.999999995")

        result = str(attotimedelta(nanoseconds=-50))
        self.assertEqual(result, "-1 day, 23:59:59.99999995")

        result = str(attotimedelta(nanoseconds=Decimal("-0.001")))
        self.assertEqual(result, "-1 day, 23:59:59.999999999999")

        result = str(
            attotimedelta(
                days=1,
                seconds=5,
                microseconds=5,
                milliseconds=5,
                minutes=50,
                hours=5,
                weeks=1,
                nanoseconds=5,
            )
        )
        self.assertEqual(result, "8 days, 5:50:05.005005005")

        result = str(
            attotimedelta(
                days=-1,
                seconds=-5,
                microseconds=-5,
                milliseconds=-5,
                minutes=-50,
                hours=-5,
                weeks=-1,
                nanoseconds=-5,
            )
        )
        self.assertEqual(result, "-9 days, 18:09:54.994994995")

    def test_repr(self):
        result = repr(attotimedelta(days=1))
        self.assertEqual(result, "attotime.objects.attotimedelta(1)")

        result = repr(attotimedelta(days=5))
        self.assertEqual(result, "attotime.objects.attotimedelta(5)")

        result = repr(attotimedelta(days=50))
        self.assertEqual(result, "attotime.objects.attotimedelta(50)")

        result = repr(attotimedelta(days=-1))
        self.assertEqual(result, "attotime.objects.attotimedelta(-1)")

        result = repr(attotimedelta(days=-5))
        self.assertEqual(result, "attotime.objects.attotimedelta(-5)")

        result = repr(attotimedelta(days=-50))
        self.assertEqual(result, "attotime.objects.attotimedelta(-50)")

        result = repr(attotimedelta(seconds=1))
        self.assertEqual(result, "attotime.objects.attotimedelta(0, 1)")

        result = repr(attotimedelta(seconds=5))
        self.assertEqual(result, "attotime.objects.attotimedelta(0, 5)")

        result = repr(attotimedelta(seconds=50))
        self.assertEqual(result, "attotime.objects.attotimedelta(0, 50)")

        result = repr(attotimedelta(seconds=-1))
        self.assertEqual(result, "attotime.objects.attotimedelta(-1, 86399)")

        result = repr(attotimedelta(seconds=-5))
        self.assertEqual(result, "attotime.objects.attotimedelta(-1, 86395)")

        result = repr(attotimedelta(seconds=-50))
        self.assertEqual(result, "attotime.objects.attotimedelta(-1, 86350)")

        result = repr(attotimedelta(microseconds=1))
        self.assertEqual(result, "attotime.objects.attotimedelta(0, 0, 1)")

        result = repr(attotimedelta(microseconds=5))
        self.assertEqual(result, "attotime.objects.attotimedelta(0, 0, 5)")

        result = repr(attotimedelta(microseconds=50))
        self.assertEqual(result, "attotime.objects.attotimedelta(0, 0, 50)")

        result = repr(attotimedelta(microseconds=-1))
        self.assertEqual(result, "attotime.objects.attotimedelta(-1, 86399, 999999)")

        result = repr(attotimedelta(microseconds=-5))
        self.assertEqual(result, "attotime.objects.attotimedelta(-1, 86399, 999995)")

        result = repr(attotimedelta(microseconds=-50))
        self.assertEqual(result, "attotime.objects.attotimedelta(-1, 86399, 999950)")

        result = repr(attotimedelta(milliseconds=1))
        self.assertEqual(result, "attotime.objects.attotimedelta(0, 0, 1000)")

        result = repr(attotimedelta(milliseconds=5))
        self.assertEqual(result, "attotime.objects.attotimedelta(0, 0, 5000)")

        result = repr(attotimedelta(milliseconds=50))
        self.assertEqual(result, "attotime.objects.attotimedelta(0, 0, 50000)")

        result = repr(attotimedelta(milliseconds=-1))
        self.assertEqual(result, "attotime.objects.attotimedelta(-1, 86399, 999000)")

        result = repr(attotimedelta(milliseconds=-5))
        self.assertEqual(result, "attotime.objects.attotimedelta(-1, 86399, 995000)")

        result = repr(attotimedelta(milliseconds=-50))
        self.assertEqual(result, "attotime.objects.attotimedelta(-1, 86399, 950000)")

        result = repr(attotimedelta(minutes=1))
        self.assertEqual(result, "attotime.objects.attotimedelta(0, 60)")

        result = repr(attotimedelta(minutes=5))
        self.assertEqual(result, "attotime.objects.attotimedelta(0, 300)")

        result = repr(attotimedelta(minutes=50))
        self.assertEqual(result, "attotime.objects.attotimedelta(0, 3000)")

        result = repr(attotimedelta(minutes=-1))
        self.assertEqual(result, "attotime.objects.attotimedelta(-1, 86340)")

        result = repr(attotimedelta(minutes=-5))
        self.assertEqual(result, "attotime.objects.attotimedelta(-1, 86100)")

        result = repr(attotimedelta(minutes=-50))
        self.assertEqual(result, "attotime.objects.attotimedelta(-1, 83400)")

        result = repr(attotimedelta(hours=1))
        self.assertEqual(result, "attotime.objects.attotimedelta(0, 3600)")

        result = repr(attotimedelta(hours=5))
        self.assertEqual(result, "attotime.objects.attotimedelta(0, 18000)")

        result = repr(attotimedelta(hours=12))
        self.assertEqual(result, "attotime.objects.attotimedelta(0, 43200)")

        result = repr(attotimedelta(hours=-1))
        self.assertEqual(result, "attotime.objects.attotimedelta(-1, 82800)")

        result = repr(attotimedelta(hours=-5))
        self.assertEqual(result, "attotime.objects.attotimedelta(-1, 68400)")

        result = repr(attotimedelta(hours=-12))
        self.assertEqual(result, "attotime.objects.attotimedelta(-1, 43200)")

        result = repr(attotimedelta(weeks=1))
        self.assertEqual(result, "attotime.objects.attotimedelta(7)")

        result = repr(attotimedelta(weeks=5))
        self.assertEqual(result, "attotime.objects.attotimedelta(35)")

        result = repr(attotimedelta(weeks=50))
        self.assertEqual(result, "attotime.objects.attotimedelta(350)")

        result = repr(attotimedelta(weeks=-1))
        self.assertEqual(result, "attotime.objects.attotimedelta(-7)")

        result = repr(attotimedelta(weeks=-5))
        self.assertEqual(result, "attotime.objects.attotimedelta(-35)")

        result = repr(attotimedelta(weeks=-50))
        self.assertEqual(result, "attotime.objects.attotimedelta(-350)")

        result = repr(attotimedelta(nanoseconds=1))
        self.assertEqual(result, "attotime.objects.attotimedelta(0, 0, 0, 1)")

        result = repr(attotimedelta(nanoseconds=5))
        self.assertEqual(result, "attotime.objects.attotimedelta(0, 0, 0, 5)")

        result = repr(attotimedelta(nanoseconds=50))
        self.assertEqual(result, "attotime.objects.attotimedelta(0, 0, 0, 50)")

        result = repr(attotimedelta(nanoseconds=Decimal("0.001")))
        self.assertEqual(result, "attotime.objects.attotimedelta(0, 0, 0, 0.001)")

        result = repr(attotimedelta(nanoseconds=-1))
        self.assertEqual(
            result, "attotime.objects.attotimedelta(-1, 86399, 999999, 999)"
        )

        result = repr(attotimedelta(nanoseconds=-5))
        self.assertEqual(
            result, "attotime.objects.attotimedelta(-1, 86399, 999999, 995)"
        )

        result = repr(attotimedelta(nanoseconds=-50))
        self.assertEqual(
            result, "attotime.objects.attotimedelta(-1, 86399, 999999, 950)"
        )

        result = repr(attotimedelta(nanoseconds=Decimal("-0.001")))
        self.assertEqual(
            result, "attotime.objects.attotimedelta(-1, 86399, 999999, 999.999)"
        )

        result = repr(
            attotimedelta(
                days=1,
                seconds=5,
                microseconds=5,
                milliseconds=5,
                minutes=50,
                hours=5,
                weeks=1,
                nanoseconds=5,
            )
        )
        self.assertEqual(result, "attotime.objects.attotimedelta(8, 21005, 5005, 5)")

        result = repr(
            attotimedelta(
                days=-1,
                seconds=-5,
                microseconds=-5,
                milliseconds=-5,
                minutes=-50,
                hours=-5,
                weeks=-1,
                nanoseconds=-5,
            )
        )
        self.assertEqual(
            result, "attotime.objects.attotimedelta(-9, 65394, 994994, 995)"
        )

    def test_total_seconds(self):
        result = attotimedelta(days=1).total_seconds()
        self.assertEqual(result, constants.SECONDS_PER_DAY)

        result = attotimedelta(seconds=1).total_seconds()
        self.assertEqual(result, 1)

        result = attotimedelta(microseconds=1).total_seconds()
        self.assertEqual(result, Decimal("1e-6"))

        result = attotimedelta(milliseconds=1).total_seconds()
        self.assertEqual(result, Decimal("1e-3"))

        result = attotimedelta(minutes=1).total_seconds()
        self.assertEqual(result, constants.SECONDS_PER_MINUTE)

        result = attotimedelta(hours=1).total_seconds()
        self.assertEqual(result, constants.SECONDS_PER_HOUR)

        result = attotimedelta(weeks=1).total_seconds()
        self.assertEqual(result, constants.SECONDS_PER_WEEK)

        result = attotimedelta(nanoseconds=1).total_seconds()
        self.assertEqual(result, Decimal("1e-9"))

        result = attotimedelta(days=-1).total_seconds()
        self.assertEqual(result, -constants.SECONDS_PER_DAY)

        result = attotimedelta(seconds=-1).total_seconds()
        self.assertEqual(result, -1)

        result = attotimedelta(microseconds=-1).total_seconds()
        self.assertEqual(result, -Decimal("1e-6"))

        result = attotimedelta(milliseconds=-1).total_seconds()
        self.assertEqual(result, -Decimal("1e-3"))

        result = attotimedelta(minutes=-1).total_seconds()
        self.assertEqual(result, -constants.SECONDS_PER_MINUTE)

        result = attotimedelta(hours=-1).total_seconds()
        self.assertEqual(result, -constants.SECONDS_PER_HOUR)

        result = attotimedelta(weeks=-1).total_seconds()
        self.assertEqual(result, -constants.SECONDS_PER_WEEK)

        result = attotimedelta(nanoseconds=-1).total_seconds()
        self.assertEqual(result, -Decimal("1e-9"))

        result = attotimedelta(
            days=1,
            seconds=1,
            microseconds=1,
            milliseconds=1,
            minutes=1,
            hours=1,
            weeks=1,
            nanoseconds=Decimal(3.14159),
        ).total_seconds()
        self.assertEqual(
            result,
            constants.SECONDS_PER_DAY
            + 1
            + Decimal("1e-6")
            + Decimal("1e-3")
            + constants.SECONDS_PER_MINUTE
            + constants.SECONDS_PER_HOUR
            + constants.SECONDS_PER_WEEK
            + (Decimal(3.14159) / constants.NANOSECONDS_PER_SECOND),
        )

        result = attotimedelta(
            days=1,
            seconds=-1,
            microseconds=1,
            milliseconds=-1,
            minutes=1,
            hours=-1,
            weeks=1,
            nanoseconds=Decimal(3.14159),
        ).total_seconds()
        self.assertEqual(
            result,
            constants.SECONDS_PER_DAY
            - 1
            + Decimal("1e-6")
            - Decimal("1e-3")
            + constants.SECONDS_PER_MINUTE
            - constants.SECONDS_PER_HOUR
            + constants.SECONDS_PER_WEEK
            + (Decimal(3.14159) / constants.NANOSECONDS_PER_SECOND),
        )

    def test_total_nanoseconds(self):
        result = attotimedelta(days=1).total_nanoseconds()
        self.assertEqual(result, constants.NANOSECONDS_PER_DAY)

        result = attotimedelta(seconds=1).total_nanoseconds()
        self.assertEqual(result, constants.NANOSECONDS_PER_SECOND)

        result = attotimedelta(microseconds=1).total_nanoseconds()
        self.assertEqual(result, constants.NANOSECONDS_PER_MICROSECOND)

        result = attotimedelta(milliseconds=1).total_nanoseconds()
        self.assertEqual(result, constants.NANOSECONDS_PER_MILLISECOND)

        result = attotimedelta(minutes=1).total_nanoseconds()
        self.assertEqual(result, constants.NANOSECONDS_PER_MINUTE)

        result = attotimedelta(hours=1).total_nanoseconds()
        self.assertEqual(result, constants.NANOSECONDS_PER_HOUR)

        result = attotimedelta(weeks=1).total_nanoseconds()
        self.assertEqual(result, constants.NANOSECONDS_PER_WEEK)

        result = attotimedelta(nanoseconds=1).total_nanoseconds()
        self.assertEqual(result, 1)

        result = attotimedelta(days=-1).total_nanoseconds()
        self.assertEqual(result, -constants.NANOSECONDS_PER_DAY)

        result = attotimedelta(seconds=-1).total_nanoseconds()
        self.assertEqual(result, -constants.NANOSECONDS_PER_SECOND)

        result = attotimedelta(microseconds=-1).total_nanoseconds()
        self.assertEqual(result, -constants.NANOSECONDS_PER_MICROSECOND)

        result = attotimedelta(milliseconds=-1).total_nanoseconds()
        self.assertEqual(result, -constants.NANOSECONDS_PER_MILLISECOND)

        result = attotimedelta(minutes=-1).total_nanoseconds()
        self.assertEqual(result, -constants.NANOSECONDS_PER_MINUTE)

        result = attotimedelta(hours=-1).total_nanoseconds()
        self.assertEqual(result, -constants.NANOSECONDS_PER_HOUR)

        result = attotimedelta(weeks=-1).total_nanoseconds()
        self.assertEqual(result, -constants.NANOSECONDS_PER_WEEK)

        result = attotimedelta(nanoseconds=-1).total_nanoseconds()
        self.assertEqual(result, -1)

        result = attotimedelta(
            days=1,
            seconds=1,
            microseconds=1,
            milliseconds=1,
            minutes=1,
            hours=1,
            weeks=1,
            nanoseconds=Decimal(3.14159),
        ).total_nanoseconds()
        self.assertEqual(
            result,
            constants.NANOSECONDS_PER_DAY
            + constants.NANOSECONDS_PER_SECOND
            + constants.NANOSECONDS_PER_MICROSECOND
            + constants.NANOSECONDS_PER_MILLISECOND
            + constants.NANOSECONDS_PER_MINUTE
            + constants.NANOSECONDS_PER_HOUR
            + constants.NANOSECONDS_PER_WEEK
            + Decimal(3.14159),
        )

        result = attotimedelta(
            days=1,
            seconds=-1,
            microseconds=1,
            milliseconds=-1,
            minutes=1,
            hours=-1,
            weeks=1,
            nanoseconds=Decimal(3.14159),
        ).total_nanoseconds()
        self.assertEqual(
            result,
            constants.NANOSECONDS_PER_DAY
            - constants.NANOSECONDS_PER_SECOND
            + constants.NANOSECONDS_PER_MICROSECOND
            - constants.NANOSECONDS_PER_MILLISECOND
            + constants.NANOSECONDS_PER_MINUTE
            - constants.NANOSECONDS_PER_HOUR
            + constants.NANOSECONDS_PER_WEEK
            + Decimal(3.14159),
        )

    def test_as_nanoseconds(self):
        result = attotimedelta._as_nanoseconds(days=1)
        self.assertEqual(result, constants.NANOSECONDS_PER_DAY)

        result = attotimedelta._as_nanoseconds(seconds=1)
        self.assertEqual(result, constants.NANOSECONDS_PER_SECOND)

        result = attotimedelta._as_nanoseconds(microseconds=1)
        self.assertEqual(result, constants.NANOSECONDS_PER_MICROSECOND)

        result = attotimedelta._as_nanoseconds(milliseconds=1)
        self.assertEqual(result, constants.NANOSECONDS_PER_MILLISECOND)

        result = attotimedelta._as_nanoseconds(minutes=1)
        self.assertEqual(result, constants.NANOSECONDS_PER_MINUTE)

        result = attotimedelta._as_nanoseconds(hours=1)
        self.assertEqual(result, constants.NANOSECONDS_PER_HOUR)

        result = attotimedelta._as_nanoseconds(weeks=1)
        self.assertEqual(result, constants.NANOSECONDS_PER_WEEK)

        result = attotimedelta._as_nanoseconds(nanoseconds=1)
        self.assertEqual(result, 1)

        result = attotimedelta._as_nanoseconds(days=-1)
        self.assertEqual(result, -constants.NANOSECONDS_PER_DAY)

        result = attotimedelta._as_nanoseconds(seconds=-1)
        self.assertEqual(result, -constants.NANOSECONDS_PER_SECOND)

        result = attotimedelta._as_nanoseconds(microseconds=-1)
        self.assertEqual(result, -constants.NANOSECONDS_PER_MICROSECOND)

        result = attotimedelta._as_nanoseconds(milliseconds=-1)
        self.assertEqual(result, -constants.NANOSECONDS_PER_MILLISECOND)

        result = attotimedelta._as_nanoseconds(minutes=-1)
        self.assertEqual(result, -constants.NANOSECONDS_PER_MINUTE)

        result = attotimedelta._as_nanoseconds(hours=-1)
        self.assertEqual(result, -constants.NANOSECONDS_PER_HOUR)

        result = attotimedelta._as_nanoseconds(weeks=-1)
        self.assertEqual(result, -constants.NANOSECONDS_PER_WEEK)

        result = attotimedelta._as_nanoseconds(nanoseconds=-1)
        self.assertEqual(result, -1)

        result = attotimedelta._as_nanoseconds(
            days=1,
            seconds=2,
            microseconds=3,
            milliseconds=4,
            minutes=5,
            hours=6,
            weeks=7,
            nanoseconds=8,
        )
        self.assertEqual(
            result,
            constants.NANOSECONDS_PER_DAY
            + 2 * constants.NANOSECONDS_PER_SECOND
            + 3 * constants.NANOSECONDS_PER_MICROSECOND
            + 4 * constants.NANOSECONDS_PER_MILLISECOND
            + 5 * constants.NANOSECONDS_PER_MINUTE
            + 6 * constants.NANOSECONDS_PER_HOUR
            + 7 * constants.NANOSECONDS_PER_WEEK
            + 8,
        )

        result = attotimedelta._as_nanoseconds(
            days=-1,
            seconds=-2,
            microseconds=-3,
            milliseconds=-4,
            minutes=-5,
            hours=-6,
            weeks=-7,
            nanoseconds=-8,
        )
        self.assertEqual(
            result,
            -(
                constants.NANOSECONDS_PER_DAY
                + 2 * constants.NANOSECONDS_PER_SECOND
                + 3 * constants.NANOSECONDS_PER_MICROSECOND
                + 4 * constants.NANOSECONDS_PER_MILLISECOND
                + 5 * constants.NANOSECONDS_PER_MINUTE
                + 6 * constants.NANOSECONDS_PER_HOUR
                + 7 * constants.NANOSECONDS_PER_WEEK
                + 8
            ),
        )

    def test_reduce_nanoseconds_int(self):
        result = attotimedelta._reduce_nanoseconds(1)
        self.assertEqual(result, (0, 0, 0, 0, 0, 0, 0, 1))

        # One microsecond
        result = attotimedelta._reduce_nanoseconds(1 * Decimal("1e3"))
        self.assertEqual(result, (0, 0, 1, 0, 0, 0, 0, 0))

        # One millisecond
        result = attotimedelta._reduce_nanoseconds(1 * Decimal("1e6"))
        self.assertEqual(result, (0, 0, 0, 1, 0, 0, 0, 0))

        # One second
        result = attotimedelta._reduce_nanoseconds(1 * Decimal("1e9"))
        self.assertEqual(result, (0, 1, 0, 0, 0, 0, 0, 0))

        # One minute
        result = attotimedelta._reduce_nanoseconds(1 * Decimal("1e9") * 60)
        self.assertEqual(result, (0, 0, 0, 0, 1, 0, 0, 0))

        # One hour
        result = attotimedelta._reduce_nanoseconds(1 * Decimal("1e9") * 60 * 60)
        self.assertEqual(result, (0, 0, 0, 0, 0, 1, 0, 0))

        # One day
        result = attotimedelta._reduce_nanoseconds(1 * Decimal("1e9") * 60 * 60 * 24)
        self.assertEqual(result, (1, 0, 0, 0, 0, 0, 0, 0))

        # One week
        result = attotimedelta._reduce_nanoseconds(
            1 * Decimal("1e9") * 60 * 60 * 24 * 7
        )
        self.assertEqual(result, (0, 0, 0, 0, 0, 0, 1, 0))

        result = attotimedelta._reduce_nanoseconds(-1)
        self.assertEqual(result, (0, 0, 0, 0, 0, 0, 0, -1))

        # One microsecond
        result = attotimedelta._reduce_nanoseconds(-1 * Decimal("1e3"))
        self.assertEqual(result, (0, 0, -1, 0, 0, 0, 0, 0))

        # One millisecond
        result = attotimedelta._reduce_nanoseconds(-1 * Decimal("1e6"))
        self.assertEqual(result, (0, 0, 0, -1, 0, 0, 0, 0))

        # One second
        result = attotimedelta._reduce_nanoseconds(-1 * Decimal("1e9"))
        self.assertEqual(result, (0, -1, 0, 0, 0, 0, 0, 0))

        # One minute
        result = attotimedelta._reduce_nanoseconds(-1 * Decimal("1e9") * 60)
        self.assertEqual(result, (0, 0, 0, 0, -1, 0, 0, 0))

        # One hour
        result = attotimedelta._reduce_nanoseconds(-1 * Decimal("1e9") * 60 * 60)
        self.assertEqual(result, (0, 0, 0, 0, 0, -1, 0, 0))

        # One day
        result = attotimedelta._reduce_nanoseconds(-1 * Decimal("1e9") * 60 * 60 * 24)
        self.assertEqual(result, (-1, 0, 0, 0, 0, 0, 0, 0))

        # One week
        result = attotimedelta._reduce_nanoseconds(
            -1 * Decimal("1e9") * 60 * 60 * 24 * 7
        )
        self.assertEqual(result, (0, 0, 0, 0, 0, 0, -1, 0))

    def test_reduce_nanoseconds_decimal(self):
        result = attotimedelta._reduce_nanoseconds(Decimal("1.123456789"))
        self.assertEqual(result, (0, 0, 0, 0, 0, 0, 0, Decimal("1.123456789")))

        result = attotimedelta._reduce_nanoseconds(Decimal("-1.123456789"))
        self.assertEqual(result, (0, 0, 0, 0, 0, 0, 0, Decimal("-1.123456789")))



# -*- coding: utf-8 -*-

# Copyright (c) 2021, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

from __future__ import absolute_import

import datetime
from decimal import Decimal, getcontext

from attotime import constants, util


class attotime(object):
    def __init__(
        self, hour=0, minute=0, second=0, microsecond=0, nanosecond=0, tzinfo=None
    ):
        # Everything is stored as an integer, except for nanoseconds, which is stored as a Decimal
        if not isinstance(hour, int):
            raise TypeError("hour must be int")

        if not isinstance(minute, int):
            raise TypeError("minute must be int")

        if not isinstance(second, int):
            raise TypeError("second must be int")

        if not isinstance(microsecond, int):
            raise TypeError("microsecond must be int")

        if type(nanosecond) not in (int, float, Decimal):
            raise TypeError("nanosecond must be int, float, or Decimal")

        if nanosecond < 0 or nanosecond >= constants.NANOSECONDS_PER_MICROSECOND:
            raise ValueError(
                "nanosecond must be in 0..{0}".format(
                    constants.NANOSECONDS_PER_MICROSECOND
                )
            )

        # Create the native time object
        self._native_time = datetime.time(
            hour=hour,
            minute=minute,
            second=second,
            microsecond=microsecond,
        )

        self._nanosecond = nanosecond

        self._tzinfo = tzinfo

    @property
    def hour(self):
        return self._native_time.hour

    @property
    def minute(self):
        return self._native_time.minute

    @property
    def second(self):
        return self._native_time.second

    @property
    def microsecond(self):
        return self._native_time.microsecond

    @property
    def nanosecond(self):
        return self._nanosecond

    @property
    def tzinfo(self):
        return self._tzinfo

    def replace(
        self,
        hour=None,
        minute=None,
        second=None,
        microsecond=None,
        nanosecond=None,
        tzinfo=True,
    ):
        if hour is None:
            hour = self.hour

        if minute is None:
            minute = self.minute

        if second is None:
            second = self.second

        if microsecond is None:
            microsecond = self.microsecond

        if nanosecond is None:
            nanosecond = self.nanosecond

        # Use True to support clearing tzinfo by setting tzinfo=None
        # https://github.com/python/cpython/blob/master/Lib/datetime.py#L1315
        if tzinfo is True:
            tzinfo = self.tzinfo

        return attotime(
            hour=hour,
            minute=minute,
            second=second,
            microsecond=microsecond,
            nanosecond=nanosecond,
            tzinfo=tzinfo,
        )

    def isoformat(self):
        # Returns a string of the form HH:MM:SS[.mmmmmm][+HH:MM]
        combined_seconds = (
            self.second
            + (self.microsecond / constants.MICROSECONDS_PER_SECOND)
            + (self.nanosecond / constants.NANOSECONDS_PER_SECOND)
        )

        if combined_seconds == 0:
            time_str = "{0:02d}:{1:02d}:00".format(self.hour, self.minute)
        elif combined_seconds < 10:
            time_str = "{0:02d}:{1:02d}:0{2}".format(
                self.hour, self.minute, util.decimal_stringify(combined_seconds)
            )
        else:
            time_str = "{0:02d}:{1:02d}:{2}".format(
                self.hour, self.minute, util.decimal_stringify(combined_seconds)
            )

        if self.utcoffset() is not None:
            offset = self.utcoffset()

            # Build a +HH:MM string
            if offset.days < 0:
                sign_str = "-"
                offset = -offset  # pylint: disable=invalid-unary-operand-type
            else:
                sign_str = "+"

            offset_hours, remaining_seconds = getcontext().divmod(
                Decimal(offset.total_seconds()), constants.SECONDS_PER_HOUR
            )
            offset_minutes, offset_seconds = getcontext().divmod(
                remaining_seconds, constants.SECONDS_PER_MINUTE
            )

            if offset_seconds == 0:
                offset_str = "{0}{1:02d}:{2:02d}".format(
                    sign_str, int(offset_hours), int(offset_minutes)
                )
            else:
                offset_str = "{0}{1:02d}:{2:02d}:{3:02d}".format(
                    sign_str,
                    int(offset_hours),
                    int(offset_minutes),
                    int(offset_seconds),
                )

            time_str += offset_str

        return time_str

    def strftime(self, formatstr):
        replacement_dict = {}

        # Build the dictionary of replacements
        for directive in [
            "%a",
            "%A",
            "%w",
            "%d",
            "%b",
            "%B",
            "%m",
            "%y",
            "%Y",
            "%H",
            "%I",
            "%p",
            "%M",
            "%S",
            "%f",
            "%z",
            "%Z",
            "%j",
            "%U",
            "%W",
            "%c",
            "%x",
            "%X",
        ]:
            replacement_dict[directive] = self._native_time.strftime(directive)

        # Build a native datetime with tzinfo so we can get timezone replacements
        if self._tzinfo is not None:
            tz_datetime = datetime.datetime(1900, 1, 1, tzinfo=self._tzinfo)

            replacement_dict["%z"] = tz_datetime.strftime("%z")
            replacement_dict["%Z"] = tz_datetime.strftime("%Z")
        else:
            replacement_dict["%z"] = ""
            replacement_dict["%Z"] = ""

        # Build the picosecond, attosecond, and yoctosecond replacements
        nanosecond_string = util.decimal_stringify(self._nanosecond)

        if "." in nanosecond_string:
            stop_index = nanosecond_string.index(".")

            left_padding = 3 - stop_index
            right_padding = 18 - len(nanosecond_string[stop_index:])
        else:
            left_padding = 3 - len(nanosecond_string)
            right_padding = 18 - left_padding - len(nanosecond_string)

        nanosecond_string = "0" * left_padding + nanosecond_string + "0" * right_padding
        nanosecond_string = nanosecond_string.replace(".", "")

        replacement_dict["%o"] = nanosecond_string[0:6]  # picoseconds
        replacement_dict["%q"] = nanosecond_string[6:12]  # attoseconds
        replacement_dict["%v"] = nanosecond_string[12:18]  # yoctoseconds

        replacement_dict["%%"] = "%"

        return util.multiple_replace(formatstr, replacement_dict)

    def utcoffset(self):
        if self.tzinfo is None:
            return None

        return self.tzinfo.utcoffset(None)

    def dst(self):
        if self.tzinfo is None:
            return None

        return self.tzinfo.dst(None)

    def tzname(self):
        if self.tzinfo is None:
            return None

        return self.tzinfo.tzname(self.tzinfo)

    def __eq__(self, y):
        # Return the result of x == y, where self is x
        if isinstance(y, self.__class__):
            if self.tzinfo is None and y.tzinfo is None:
                # Do a naive comparison
                return (
                    self._native_time == y._native_time
                    and self._nanosecond == y._nanosecond
                )
            elif self.tzinfo is not None and y.tzinfo is not None:
                # Do a tz aware comparison, since there is currently not a nanosecond aware
                # tzinfo implementation, just compare each part separately
                return (
                    self._native_time.replace(tzinfo=self.tzinfo)
                    == y._native_time.replace(tzinfo=y.tzinfo)
                    and self._nanosecond == y._nanosecond
                )
            else:
                raise TypeError("can't compare offset-naive and offset-aware times")

        return False

    def __ne__(self, y):
        # Return the result of x != y, where self is x
        return not self.__eq__(y)

    def __gt__(self, y):
        # Return the result of x > y, where self is x
        if isinstance(y, self.__class__):
            if self.tzinfo is None and y.tzinfo is None:
                # Do a naive comparison
                return (self._native_time > y._native_time) or (
                    self._native_time == y._native_time
                    and self._nanosecond > y._nanosecond
                )
            elif self.tzinfo is not None and y.tzinfo is not None:
                # Do a tz aware comparison, since there is currently not a nanosecond aware
                # tzinfo implementation, just compare each part separately
                return (
                    self._native_time.replace(tzinfo=self.tzinfo)
                    > y._native_time.replace(tzinfo=y.tzinfo)
                    and self._nanosecond == y._nanosecond
                ) or (
                    self._native_time.replace(tzinfo=self.tzinfo)
                    == y._native_time.replace(tzinfo=y.tzinfo)
                    and self._nanosecond > y._nanosecond
                )
            else:
                raise TypeError("can't compare offset-naive and offset-aware times")
        else:
            raise TypeError(
                "can't compare {0} to {1}".format(type(self).__name__, type(y).__name__)
            )

    def __ge__(self, y):
        # Return the result of x >= y, where self is x
        if isinstance(y, self.__class__):
            return (self > y) or (self == y)
        else:
            raise TypeError(
                "can't compare {0} to {1}".format(type(self).__name__, type(y).__name__)
            )

    def __lt__(self, y):
        # Return the result of x < y, where self is x
        if isinstance(y, self.__class__):
            if self.tzinfo is None and y.tzinfo is None:
                # Do a naive comparison
                return (self._native_time < y._native_time) or (
                    self._native_time == y._native_time
                    and self._nanosecond < y._nanosecond
                )
            elif self.tzinfo is not None and y.tzinfo is not None:
                # Do a tz aware comparison, since there is currently not a nanosecond aware
                # tzinfo implementation, just compare each part separately
                return (
                    self._native_time.replace(tzinfo=self.tzinfo)
                    < y._native_time.replace(tzinfo=y.tzinfo)
                    and self._nanosecond == y._nanosecond
                ) or (
                    self._native_time.replace(tzinfo=self.tzinfo)
                    == y._native_time.replace(tzinfo=y.tzinfo)
                    and self._nanosecond < y._nanosecond
                )
            else:
                raise TypeError("can't compare offset-naive and offset-aware times")
        else:
            raise TypeError(
                "can't compare {0} to {1}".format(type(self).__name__, type(y).__name__)
            )

    def __le__(self, y):
        # Return the result of x <= y, where self is x
        if isinstance(y, self.__class__):
            return (self < y) or (self == y)
        else:
            raise TypeError(
                "can't compare {0} to {1}".format(type(self).__name__, type(y).__name__)
            )

    def __str__(self):
        return self.isoformat()

    def __repr__(self):
        if self._tzinfo is None:
            return "{0}({1}, {2}, {3}, {4}, {5})".format(
                self.__class__.__module__,
                self.hour,
                self.minute,
                self.second,
                self.microsecond,
                str(self.nanosecond),
            )

        return "{0}({1}, {2}, {3}, {4}, {5}, {6})".format(
            self.__class__.__module__,
            self.hour,
            self.minute,
            self.second,
            self.microsecond,
            str(self.nanosecond),
            str(self._tzinfo),
        )

    def __format__(self, formatstr):
        return self.strftime(formatstr)



# -*- coding: utf-8 -*-

# Copyright (c) 2021, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

# Expose the API
from attotime.objects.attodatetime import attodatetime
from attotime.objects.attotime import attotime
from attotime.objects.attotimedelta import attotimedelta
from attotime.version import __version__



# -*- coding: utf-8 -*-

# Copyright (c) 2021, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

from __future__ import absolute_import

import datetime
import math
from decimal import Decimal, getcontext

from attotime import constants, util


class attotimedelta(object):
    def __init__(
        self,
        days=0,
        seconds=0,
        microseconds=0,
        milliseconds=0,
        minutes=0,
        hours=0,
        weeks=0,
        nanoseconds=0,
    ):
        if type(days) not in (int, float, Decimal):
            raise TypeError("days must be int, float, or Decimal")

        if type(seconds) not in (int, float, Decimal):
            raise TypeError("seconds must be int, float, or Decimal")

        if type(microseconds) not in (int, float, Decimal):
            raise TypeError("microseconds must be int, float, or Decimal")

        if type(milliseconds) not in (int, float, Decimal):
            raise TypeError("milliseconds must be int, float, or Decimal")

        if type(minutes) not in (int, float, Decimal):
            raise TypeError("minutes must be int, float, or Decimal")

        if type(hours) not in (int, float, Decimal):
            raise TypeError("hours must be int, float, or Decimal")

        if type(weeks) not in (int, float, Decimal):
            raise TypeError("weeks must be int, float, or Decimal")

        if type(nanoseconds) not in (int, float, Decimal):
            raise TypeError("nanoseconds must be int, float, or Decimal")

        # Convert everything to nanoseconds
        total_nanoseconds = self._as_nanoseconds(
            days=days,
            seconds=seconds,
            microseconds=microseconds,
            milliseconds=milliseconds,
            minutes=minutes,
            hours=hours,
            weeks=weeks,
            nanoseconds=nanoseconds,
        )

        results_tuple = self._reduce_nanoseconds(total_nanoseconds)

        days = results_tuple[0]
        seconds = results_tuple[1]
        microseconds = results_tuple[2]
        milliseconds = results_tuple[3]
        minutes = results_tuple[4]
        hours = results_tuple[5]
        weeks = results_tuple[6]
        nanosesconds = results_tuple[7]

        # Create the native timedelta
        self._native_timedelta = datetime.timedelta(
            days=int(days),
            seconds=int(seconds),
            microseconds=int(microseconds),
            milliseconds=int(milliseconds),
            minutes=int(minutes),
            hours=int(hours),
            weeks=int(weeks),
        )

        # Store the nanoseconds
        self._nanoseconds = Decimal(nanosesconds)

    @property
    def days(self):
        return self._native_timedelta.days

    @property
    def seconds(self):
        return self._native_timedelta.seconds

    @property
    def microseconds(self):
        return self._native_timedelta.microseconds

    @property
    def nanoseconds(self):
        return self._nanoseconds

    def __add__(self, y):
        # Return the result of x + y, where self is x
        x_nanoseconds = self.total_nanoseconds()
        y_nanoseconds = y.total_nanoseconds()

        result_tuple = self._reduce_nanoseconds(x_nanoseconds + y_nanoseconds)

        return attotimedelta(
            days=result_tuple[0],
            seconds=result_tuple[1],
            microseconds=result_tuple[2],
            milliseconds=result_tuple[3],
            minutes=result_tuple[4],
            hours=result_tuple[5],
            weeks=result_tuple[6],
            nanoseconds=result_tuple[7],
        )

    def __sub__(self, y):
        # Return the result of x - y, where self is x
        return self + -y

    def __mul__(self, y):
        # Return the result of x * y, where self is x and y is an integer, float, or Decimal
        x_nanoseconds = self.total_nanoseconds()

        result_tuple = self._reduce_nanoseconds(x_nanoseconds * Decimal(y))

        return attotimedelta(
            days=result_tuple[0],
            seconds=result_tuple[1],
            microseconds=result_tuple[2],
            milliseconds=result_tuple[3],
            minutes=result_tuple[4],
            hours=result_tuple[5],
            weeks=result_tuple[6],
            nanoseconds=result_tuple[7],
        )

    def __rmul__(self, y):
        # Return the result of y * x, where self ix x and y is an integer, float, or Decimal
        return self.__mul__(y)

    def __div__(self, y):
        # Return the result of x / y, where self is x and y is an integer, float, or Decimal
        x_nanoseconds = self.total_nanoseconds()

        result_tuple = self._reduce_nanoseconds(x_nanoseconds / Decimal(y))

        return attotimedelta(
            days=result_tuple[0],
            seconds=result_tuple[1],
            microseconds=result_tuple[2],
            milliseconds=result_tuple[3],
            minutes=result_tuple[4],
            hours=result_tuple[5],
            weeks=result_tuple[6],
            nanoseconds=result_tuple[7],
        )

    def __truediv__(self, y):
        # Python3 division
        return self.__div__(y)

    def __floordiv__(self, y):
        # Return the result of x // y, where self is x and y is an integer, float, or Decimal
        x_nanoseconds = self.total_nanoseconds()

        result_tuple = self._reduce_nanoseconds(x_nanoseconds // Decimal(y))

        return attotimedelta(
            days=result_tuple[0],
            seconds=result_tuple[1],
            microseconds=result_tuple[2],
            milliseconds=result_tuple[3],
            minutes=result_tuple[4],
            hours=result_tuple[5],
            weeks=result_tuple[6],
            nanoseconds=result_tuple[7],
        )

    def __eq__(self, y):
        # Return the result of x == y, where self is x
        if isinstance(y, self.__class__):
            return self.total_nanoseconds() == y.total_nanoseconds()

        # Other class doesn't match
        return False

    def __ne__(self, y):
        # Return the result of x != y, where self is x
        return not self.__eq__(y)

    def __gt__(self, y):
        # Return the result of x > y, where self is x
        if isinstance(y, self.__class__):
            return self.total_nanoseconds() > y.total_nanoseconds()
        else:
            raise TypeError(
                "can't compare {0} to {1}".format(type(self).__name__, type(y).__name__)
            )

    def __ge__(self, y):
        # Return the result of x >= y, where self is x
        if isinstance(y, self.__class__):
            return self.total_nanoseconds() >= y.total_nanoseconds()
        else:
            raise TypeError(
                "can't compare {0} to {1}".format(type(self).__name__, type(y).__name__)
            )

    def __lt__(self, y):
        # Return the result of x < y, where self is x
        if isinstance(y, self.__class__):
            return self.total_nanoseconds() < y.total_nanoseconds()
        else:
            raise TypeError(
                "can't compare {0} to {1}".format(type(self).__name__, type(y).__name__)
            )

    def __le__(self, y):
        # Return the result of x <= y, where self is x
        if isinstance(y, self.__class__):
            return self.total_nanoseconds() <= y.total_nanoseconds()
        else:
            raise TypeError(
                "can't compare {0} to {1}".format(type(self).__name__, type(y).__name__)
            )

    def __pos__(self):
        # Return the result of +x, where self is x
        #
        # https://stackoverflow.com/questions/16819023/whats-the-purpose-of-the-pos-unary-operator-in-python
        return attotimedelta(nanoseconds=+self.total_nanoseconds())

    def __neg__(self):
        # Return the result of -x, where self is x
        return attotimedelta(nanoseconds=-self.total_nanoseconds())

    def __abs__(self):
        # Return the result of abs(x), where self is x
        return attotimedelta(nanoseconds=abs(self.total_nanoseconds()))

    def __str__(self):
        # Returns a string in the form [D day[s], ][H]H:MM:SS[.UUUUUU], where D is negative for negative t.
        nanoseconds = self.total_nanoseconds()

        reduced_tuple = self._reduce_nanoseconds(nanoseconds)

        if nanoseconds < 0:
            if (
                reduced_tuple[1] != 0
                or reduced_tuple[2] != 0
                or reduced_tuple[3] != 0
                or reduced_tuple[4] != 0
                or reduced_tuple[5] != 0
                or reduced_tuple[7] != 0
            ):
                # We go to a negative day, then count forward
                days = -(
                    abs(reduced_tuple[0])
                    + abs(reduced_tuple[6]) * constants.DAYS_PER_WEEK
                    + 1
                )
            else:
                days = -(
                    abs(reduced_tuple[0])
                    + abs(reduced_tuple[6]) * constants.DAYS_PER_WEEK
                )

            seconds = abs(reduced_tuple[1])
            microseconds = abs(reduced_tuple[2])
            milliseconds = abs(reduced_tuple[3])
            minutes = abs(reduced_tuple[4])
            hours = abs(reduced_tuple[5])
            nanoseconds = abs(reduced_tuple[7])

            # Now count forward
            remaining_nanoseconds = constants.NANOSECONDS_PER_DAY - (
                seconds * constants.NANOSECONDS_PER_SECOND
                + microseconds * constants.NANOSECONDS_PER_MICROSECOND
                + milliseconds * constants.NANOSECONDS_PER_MILLISECOND
                + minutes * constants.NANOSECONDS_PER_MINUTE
                + hours * constants.NANOSECONDS_PER_HOUR
                + nanoseconds
            )

            reduced_tuple = self._reduce_nanoseconds(remaining_nanoseconds)

            seconds = reduced_tuple[1]
            microseconds = reduced_tuple[2]
            milliseconds = reduced_tuple[3]
            minutes = reduced_tuple[4]
            hours = reduced_tuple[5]
            nanoseconds = reduced_tuple[7]
        else:
            days = reduced_tuple[0] + reduced_tuple[6] * constants.DAYS_PER_WEEK
            seconds = reduced_tuple[1]
            microseconds = reduced_tuple[2]
            milliseconds = reduced_tuple[3]
            minutes = reduced_tuple[4]
            hours = reduced_tuple[5]
            nanoseconds = reduced_tuple[7]

        # Combine the seconds
        combined_seconds = (
            seconds
            + (microseconds / constants.MICROSECONDS_PER_SECOND)
            + (milliseconds / constants.MILLISECONDS_PER_SECOND)
            + (nanoseconds / constants.NANOSECONDS_PER_SECOND)
        )

        # Build the strings
        if days == 0:
            days_str = ""
        elif abs(days) > 1:
            days_str = str(days) + " days"
        else:
            days_str = str(days) + " day"

        if combined_seconds == 0:
            time_str = str(hours) + ":" + "{0:02d}".format(minutes) + ":00"
        elif combined_seconds < 10:
            time_str = (
                str(hours)
                + ":"
                + "{0:02d}".format(minutes)
                + ":0"
                + util.decimal_stringify(combined_seconds)
            )
        else:
            time_str = (
                str(hours)
                + ":"
                + "{0:02d}".format(minutes)
                + ":"
                + util.decimal_stringify(combined_seconds)
            )

        if days_str == "":
            return time_str
        else:
            return days_str + ", " + time_str

    def __repr__(self):
        # Returns a string in the form attotime.timedelta(D[, S[, U[, N]]]), where D is negative for negative t
        nanoseconds = self.total_nanoseconds()

        reduced_tuple = self._reduce_nanoseconds(nanoseconds)

        if nanoseconds < 0:
            if (
                reduced_tuple[1] != 0
                or reduced_tuple[2] != 0
                or reduced_tuple[3] != 0
                or reduced_tuple[4] != 0
                or reduced_tuple[5] != 0
                or reduced_tuple[7] != 0
            ):
                # We go to a negative day, then count forward
                days = -(
                    abs(reduced_tuple[0])
                    + abs(reduced_tuple[6]) * constants.DAYS_PER_WEEK
                    + 1
                )
            else:
                days = -(
                    abs(reduced_tuple[0])
                    + abs(reduced_tuple[6]) * constants.DAYS_PER_WEEK
                )

            seconds = abs(reduced_tuple[1])
            microseconds = abs(reduced_tuple[2])
            milliseconds = abs(reduced_tuple[3])
            minutes = abs(reduced_tuple[4])
            hours = abs(reduced_tuple[5])
            nanoseconds = abs(reduced_tuple[7])

            # Now count forward
            remaining_nanoseconds = constants.NANOSECONDS_PER_DAY - (
                seconds * constants.NANOSECONDS_PER_SECOND
                + microseconds * constants.NANOSECONDS_PER_MICROSECOND
                + milliseconds * constants.NANOSECONDS_PER_MILLISECOND
                + minutes * constants.NANOSECONDS_PER_MINUTE
                + hours * constants.NANOSECONDS_PER_HOUR
                + nanoseconds
            )

            reduced_tuple = self._reduce_nanoseconds(remaining_nanoseconds)

            seconds = reduced_tuple[1]
            microseconds = reduced_tuple[2]
            milliseconds = reduced_tuple[3]
            minutes = reduced_tuple[4]
            hours = reduced_tuple[5]
            nanoseconds = reduced_tuple[7]
        else:
            days = reduced_tuple[0] + reduced_tuple[6] * constants.DAYS_PER_WEEK
            seconds = reduced_tuple[1]
            microseconds = reduced_tuple[2]
            milliseconds = reduced_tuple[3]
            minutes = reduced_tuple[4]
            hours = reduced_tuple[5]
            nanoseconds = reduced_tuple[7]

        combined_seconds = (
            hours * constants.SECONDS_PER_HOUR
            + minutes * constants.SECONDS_PER_MINUTE
            + seconds
        )
        combined_microseconds = (
            milliseconds * constants.MICROSECONDS_PER_MILLISECOND + microseconds
        )

        repr_string = self.__class__.__module__ + "(" + str(days)

        if combined_seconds != 0 or combined_microseconds != 0 or nanoseconds != 0:
            repr_string += ", " + str(combined_seconds)

        if combined_microseconds != 0 or nanoseconds != 0:
            repr_string += ", " + str(combined_microseconds)

        if nanoseconds != 0:
            repr_string += ", " + util.decimal_stringify(nanoseconds)

        repr_string += ")"

        return repr_string

    def total_seconds(self):
        return self.total_nanoseconds() / constants.NANOSECONDS_PER_SECOND

    def total_nanoseconds(self):
        # Converts the entire timedelta to nanoseconds
        return self._as_nanoseconds(
            days=self._native_timedelta.days,
            seconds=self._native_timedelta.seconds,
            microseconds=self._native_timedelta.microseconds,
            nanoseconds=self._nanoseconds,
        )

    @staticmethod
    def _as_nanoseconds(
        days=0,
        seconds=0,
        microseconds=0,
        milliseconds=0,
        minutes=0,
        hours=0,
        weeks=0,
        nanoseconds=0,
    ):
        return (
            Decimal(days) * constants.NANOSECONDS_PER_DAY
            + Decimal(seconds) * constants.NANOSECONDS_PER_SECOND
            + Decimal(microseconds) * constants.NANOSECONDS_PER_MICROSECOND
            + Decimal(milliseconds) * constants.NANOSECONDS_PER_MILLISECOND
            + Decimal(minutes) * constants.NANOSECONDS_PER_MINUTE
            + Decimal(hours) * constants.NANOSECONDS_PER_HOUR
            + Decimal(weeks) * constants.NANOSECONDS_PER_WEEK
            + Decimal(nanoseconds)
        )

    @staticmethod
    def _reduce_nanoseconds(nanoseconds):
        return _reduce_to_tuple(
            nanoseconds,
            constants.NANOSECONDS_PER_WEEK,
            constants.NANOSECONDS_PER_DAY,
            constants.NANOSECONDS_PER_HOUR,
            constants.NANOSECONDS_PER_MINUTE,
            constants.NANOSECONDS_PER_SECOND,
            constants.NANOSECONDS_PER_MILLISECOND,
            constants.NANOSECONDS_PER_MICROSECOND,
            1,
        )


def _reduce_to_tuple(
    value,
    week_reduction,
    day_reduction,
    hour_reduction,
    minute_reduction,
    second_reduction,
    millisecond_reduction,
    microsecond_reduction,
    nanosecond_reduction,
):
    remainder = abs(Decimal(value))

    reduced_weeks, remainder = getcontext().divmod(remainder, week_reduction)
    reduced_days, remainder = getcontext().divmod(remainder, day_reduction)
    reduced_hours, remainder = getcontext().divmod(remainder, hour_reduction)
    reduced_minutes, remainder = getcontext().divmod(remainder, minute_reduction)
    reduced_seconds, remainder = getcontext().divmod(remainder, second_reduction)
    reduced_milliseconds, remainder = getcontext().divmod(
        remainder, millisecond_reduction
    )
    reduced_microseconds, remainder = getcontext().divmod(
        remainder, microsecond_reduction
    )
    nanoseconds = remainder * nanosecond_reduction

    # Return the mess, ints for everything but nanoseconds
    return (
        int(math.copysign(reduced_days, value)),
        int(math.copysign(reduced_seconds, value)),
        int(math.copysign(reduced_microseconds, value)),
        int(math.copysign(reduced_milliseconds, value)),
        int(math.copysign(reduced_minutes, value)),
        int(math.copysign(reduced_hours, value)),
        int(math.copysign(reduced_weeks, value)),
        nanoseconds.copy_sign(Decimal(value)),
    )



# -*- coding: utf-8 -*-

# Copyright (c) 2021, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

from __future__ import absolute_import

import calendar
import datetime
import time
from decimal import Decimal, getcontext

from attotime import compat, constants, util
from attotime.objects.attotime import attotime
from attotime.objects.attotimedelta import attotimedelta
from attotime.util.strptime import (
    expand_format_fields,
    get_field_size,
    get_format_fields,
)


class attodatetime(object):
    def __init__(
        self,
        year,
        month,
        day,
        hour=0,
        minute=0,
        second=0,
        microsecond=0,
        nanosecond=0,
        tzinfo=None,
    ):
        if not isinstance(year, int):
            raise TypeError("year must be int")

        if not isinstance(month, int):
            raise TypeError("month must be int")

        if not isinstance(day, int):
            raise TypeError("day must be int")

        # Create the native date object
        self._native_date = datetime.date(year=year, month=month, day=day)

        # Create the attotime time object
        self._attotime = attotime(
            hour=hour,
            minute=minute,
            second=second,
            microsecond=microsecond,
            nanosecond=nanosecond,
            tzinfo=tzinfo,
        )

    @classmethod
    def today(cls):
        return cls.fromtimestamp(time.time())

    @classmethod
    def now(cls, tz=None):
        if tz is None:
            return cls.today()

        return cls.fromtimestamp(time.time(), tz)

    @classmethod
    def utcnow(cls):
        return cls.utcfromtimestamp(time.time())

    @classmethod
    def fromtimestamp(cls, timestamp, tz=None):
        utc_datetime = cls.utcfromtimestamp(timestamp)

        if tz is None:
            # Apply the local timezone, return as naive
            tz_delta = attotimedelta(
                seconds=util.LocalTimezone().utcoffset(utc_datetime).total_seconds()
            )
            return utc_datetime + tz_delta

        # Apply the timezone
        tz_delta = attotimedelta(
            seconds=tz.utcoffset(utc_datetime._as_datetime()).total_seconds()
        )
        return (utc_datetime + tz_delta).replace(tzinfo=tz)

    @classmethod
    def utcfromtimestamp(cls, timestamp):
        epoch_struct = time.gmtime(0)
        epoch_datetime = cls(
            year=epoch_struct[0],
            month=epoch_struct[1],
            day=epoch_struct[2],
            hour=epoch_struct[3],
            minute=epoch_struct[4],
            second=epoch_struct[5],
        )

        timestamp_delta = attotimedelta(seconds=timestamp)

        return epoch_datetime + timestamp_delta

    @classmethod
    def fromordinal(cls, ordinal):
        date = datetime.datetime.fromordinal(ordinal)

        return cls(year=date.year, month=date.month, day=date.day)

    @classmethod
    def combine(cls, date, time):
        result = cls(year=1, month=1, day=1)

        # Now replace the instance objects
        result._native_date = datetime.date(
            year=date.year, month=date.month, day=date.day
        )
        result._attotime = time

        return result

    @classmethod
    def strptime(cls, date_string, format):
        FIXED_LENGTH_FIELDS = [
            "%a",
            "%A",
            "%w",
            "%b",
            "%B",
            "%y",
            "%Y",
            "%p",
            "%x",
            "%X",
            "%%",
        ]

        if "%o" in format or "%q" in format or "%v" in format:
            # We have to process the sub-microsecond fields
            format_fields = expand_format_fields(get_format_fields(format))
            string_index = 0
            format_span_sizes = []

            for field_index in compat.range(0, len(format_fields)):
                format_field = format_fields[field_index]

                format_span_search_start = 0
                format_span_search_end = 0

                for span_size in format_span_sizes:
                    if isinstance(span_size, tuple):
                        # Search in a range
                        format_span_search_start += span_size[0]
                        format_span_search_end += span_size[1]
                    else:
                        # Exact span size is known
                        format_span_search_start += span_size
                        format_span_search_end += span_size

                if not format_field.startswith("%"):
                    # Span is fixed
                    format_span_sizes.append(len(format_field))
                elif format_field in FIXED_LENGTH_FIELDS:
                    # Span is fixed, but needs to be determined
                    span_size = -1

                    if format_field == "%a":
                        # Flip through all available weekday abbreviations
                        day_abbrs = []

                        for day_index in compat.range(0, 7):
                            day_abbrs.append(calendar.day_abbr[day_index])

                        span_size = get_field_size(
                            date_string,
                            day_abbrs,
                            format_span_search_start,
                            format_span_search_end,
                        )
                    elif format_field == "%A":
                        # Flip through all available weekdays
                        day_names = []

                        for day_index in compat.range(0, 7):
                            day_names.append(calendar.day_name[day_index])

                        span_size = get_field_size(
                            date_string,
                            day_names,
                            format_span_search_start,
                            format_span_search_end,
                        )
                    elif format_field == "%w":
                        span_size = 1
                    elif format_field == "%b":
                        # Check all available month abbreviations
                        month_abbrs = []

                        for month_index in compat.range(1, 13):
                            month_abbrs.append(calendar.month_abbr[month_index])

                        span_size = get_field_size(
                            date_string,
                            month_abbrs,
                            format_span_search_start,
                            format_span_search_end,
                        )
                    elif format_field == "%B":
                        # Check all available month names
                        month_names = []

                        for month_index in compat.range(1, 13):
                            month_names.append(calendar.month_name[month_index])

                        span_size = get_field_size(
                            date_string,
                            month_names,
                            format_span_search_start,
                            format_span_search_end,
                        )
                    elif format_field == "%y":
                        span_size = 2
                    elif format_field == "%Y":
                        span_size = 4
                    elif format_field == "%p":
                        # Taken from the Python strptime implementation
                        # https://github.com/python/cpython/blob/master/Lib/_strptime.py#L95
                        am_pm = []

                        for hour in (1, 22):
                            time_tuple = time.struct_time(
                                (1999, 3, 17, hour, 44, 55, 2, 76, 0)
                            )
                            am_pm.append(time.strftime("%p", time_tuple).lower())

                        if "" in am_pm:
                            # Some locales don't have am/pm designator
                            span_size = 0
                        else:
                            span_size = get_field_size(
                                date_string,
                                am_pm,
                                format_span_search_start,
                                format_span_search_end,
                            )
                    else:
                        # %%
                        span_size = 1

                    if span_size == -1:
                        raise ValueError("Span {0} not found".format(format_field))

                    format_span_sizes.append(span_size)
                else:
                    # Span is a range
                    if format_field in ["%d", "%m", "%H", "%I", "%M", "%S", "%U", "%W"]:
                        # Could be 1 or 2 characters
                        format_span_sizes.append((1, 2))
                    elif format_field in ["%f", "%o", "%q", "%v"]:
                        format_span_sizes.append((1, 6))
                    elif format_field == "%j":
                        format_span_sizes.append((1, 3))

            # Now calculate the size of the range spans
            used_length = 0
            unknown_span_ranges = []
            unknown_spans = []

            for span_index in compat.range(0, len(format_span_sizes)):
                span_size = format_span_sizes[span_index]

                if isinstance(span_size, tuple):
                    unknown_span_ranges.append(span_size)
                    unknown_spans.append(span_index)
                else:
                    used_length += span_size

            # Distribute remaining length
            remaining_length = len(date_string) - used_length
            unknown_span_index = 0

            for unknown_span_index in unknown_spans:
                format_span_sizes[unknown_span_index] = 0

            unknown_span_index = 0
            length_distributed = 0

            while remaining_length > 0:
                current_span_size = format_span_sizes[unknown_spans[unknown_span_index]]

                if current_span_size + 1 <= max(
                    unknown_span_ranges[unknown_span_index]
                ):
                    format_span_sizes[unknown_spans[unknown_span_index]] += 1
                    remaining_length -= 1
                    length_distributed += 1

                unknown_span_index += 1

                if unknown_span_index == len(unknown_spans):
                    if length_distributed == 0:
                        break

                    unknown_span_index = 0
                    length_distributed = 0

            # Now that we know all the lengths we can split up the date string
            python_date_string = ""
            python_format_fields = ""

            atto_date_string = ""
            atto_format_fields = ""

            string_index = 0

            for span_index in compat.range(0, len(format_span_sizes)):
                format_field = format_fields[span_index]
                span_size = format_span_sizes[span_index]

                if format_field in ["%o", "%q", "%v"]:
                    atto_date_string += date_string[
                        string_index : string_index + span_size
                    ]
                    atto_format_fields += format_field
                else:
                    python_date_string += date_string[
                        string_index : string_index + span_size
                    ]
                    python_format_fields += format_field

                string_index += span_size

            # Build the python datetime
            python_datetime = datetime.datetime.strptime(
                python_date_string, python_format_fields
            )
            atto_datetime = attodatetime(
                python_datetime.year,
                python_datetime.month,
                python_datetime.day,
                hour=python_datetime.hour,
                minute=python_datetime.minute,
                second=python_datetime.second,
                microsecond=python_datetime.microsecond,
                tzinfo=python_datetime.tzinfo,
            )

            # Build the attotimedelta
            nanoseconds_accumulator = Decimal(0)

            string_index = 0

            for span_index in compat.range(0, len(format_span_sizes)):
                format_field = format_fields[span_index]
                span_size = format_span_sizes[span_index]

                date_component = atto_date_string[
                    string_index : string_index + span_size
                ]

                if format_field == "%o":
                    nanoseconds_accumulator += (
                        Decimal(date_component) / constants.PICOSECONDS_PER_NANOSECOND
                    )
                    string_index += span_size
                elif format_field == "%q":
                    nanoseconds_accumulator += (
                        Decimal(date_component) / constants.ATTOSECONDS_PER_NANOSECOND
                    )
                    string_index += span_size
                elif format_field == "%v":
                    nanoseconds_accumulator += (
                        Decimal(date_component) / constants.YOCTOSECONDS_PER_NANOSECOND
                    )
                    string_index += span_size

            atto_delta = attotimedelta(nanoseconds=nanoseconds_accumulator)

            return atto_datetime + atto_delta
        else:
            # We can simply call the Python implementation
            python_datetime = datetime.datetime.strptime(date_string, format)

            return attodatetime(
                python_datetime.year,
                python_datetime.month,
                python_datetime.day,
                hour=python_datetime.hour,
                minute=python_datetime.minute,
                second=python_datetime.second,
                microsecond=python_datetime.microsecond,
                tzinfo=python_datetime.tzinfo,
            )

    @property
    def year(self):
        return self._native_date.year

    @property
    def month(self):
        return self._native_date.month

    @property
    def day(self):
        return self._native_date.day

    @property
    def hour(self):
        return self._attotime.hour

    @property
    def minute(self):
        return self._attotime.minute

    @property
    def second(self):
        return self._attotime.second

    @property
    def microsecond(self):
        return self._attotime.microsecond

    @property
    def nanosecond(self):
        return self._attotime.nanosecond

    @property
    def tzinfo(self):
        return self._attotime.tzinfo

    def date(self):
        return self._as_datetime().date()

    def time(self):
        return attotime(
            hour=self.hour,
            minute=self.minute,
            second=self.second,
            microsecond=self.microsecond,
            nanosecond=self.nanosecond,
        )

    def timetz(self):
        return attotime(
            hour=self.hour,
            minute=self.minute,
            second=self.second,
            microsecond=self.microsecond,
            nanosecond=self.nanosecond,
            tzinfo=self.tzinfo,
        )

    def replace(
        self,
        year=None,
        month=None,
        day=None,
        hour=None,
        minute=None,
        second=None,
        microsecond=None,
        nanosecond=None,
        tzinfo=True,
    ):
        if year is None:
            year = self.year

        if month is None:
            month = self.month

        if day is None:
            day = self.day

        if hour is None:
            hour = self.hour

        if minute is None:
            minute = self.minute

        if second is None:
            second = self.second

        if microsecond is None:
            microsecond = self.microsecond

        if nanosecond is None:
            nanosecond = self.nanosecond

        # Use True to support clearing tzinfo by setting tzinfo=None
        # https://github.com/python/cpython/blob/master/Lib/datetime.py#L1596
        if tzinfo is True:
            tzinfo = self.tzinfo

        return attodatetime(
            year=year,
            month=month,
            day=day,
            hour=hour,
            minute=minute,
            second=second,
            microsecond=microsecond,
            nanosecond=nanosecond,
            tzinfo=tzinfo,
        )

    def astimezone(self, tz):
        if self.tzinfo is None:
            raise ValueError("astimezone() cannot be applied to a naive datetime")

        if self.tzinfo is tz:
            return self

        # Convert to UTC, remove timezone info
        utc_datetime = (self - self.utcoffset()).replace(tzinfo=None)

        # Apply the timezone
        tz_delta = attotimedelta(
            seconds=tz.utcoffset(utc_datetime._as_datetime()).total_seconds()
        )
        return (utc_datetime + tz_delta).replace(tzinfo=tz)

    def utcoffset(self):
        # Offsets don't have nanosecond resolution, so build a native datetime
        native_datetime = self._as_datetime()

        native_timedelta = native_datetime.utcoffset()

        if native_timedelta is None:
            return None

        # Convert the native timedelta to an attotimedelta
        return attotimedelta(
            days=native_timedelta.days,
            seconds=native_timedelta.seconds,
            microseconds=native_timedelta.microseconds,
            nanoseconds=0,
        )

    def dst(self):
        if self.tzinfo is None:
            return None

        # Offsets don't have nanosecond resolution, so build a native datetime
        native_datetime = self._as_datetime()

        native_timedelta = native_datetime.tzinfo.dst(native_datetime)

        # Convert the native timedelta to an attotimedelta
        return attotimedelta(
            days=native_timedelta.days,
            seconds=native_timedelta.seconds,
            microseconds=native_timedelta.microseconds,
            nanoseconds=0,
        )

    def tzname(self):
        if self.tzinfo is None:
            return None

        return self.tzinfo.tzname(self.tzinfo)

    def timetuple(self):
        return self._as_datetime().timetuple()

    def utctimetuple(self):
        return self._as_datetime().utctimetuple()

    def toordinal(self):
        return self._as_datetime().toordinal()

    def weekday(self):
        return self._as_datetime().weekday()

    def isoweekday(self):
        return self._as_datetime().isoweekday()

    def isocalendar(self):
        return self._as_datetime().isocalendar()

    def isoformat(self, separator="T"):
        return separator.join(
            [self._as_datetime().date().isoformat(), self._attotime.isoformat()]
        )

    def ctime(self):
        return self._as_datetime().ctime()

    def strftime(self, formatstr):
        replacement_dict = {}

        # Build the dictionary of replacements
        for directive in [
            "%a",
            "%A",
            "%w",
            "%d",
            "%b",
            "%B",
            "%m",
            "%y",
            "%Y",
            "%j",
            "%U",
            "%W",
            "%x",
        ]:
            replacement_dict[directive] = self._native_date.strftime(directive)

        native_time = self._attotime._native_time

        for directive in ["%H", "%I", "%p", "%M", "%S", "%f", "%z", "%Z", "%X"]:
            replacement_dict[directive] = native_time.strftime(directive)

        # Build a native datetime with tzinfo so we can get timezone replacements
        if self._attotime.tzinfo is not None:
            tz_datetime = datetime.datetime(1900, 1, 1, tzinfo=self._attotime.tzinfo)

            replacement_dict["%z"] = tz_datetime.strftime("%z")
            replacement_dict["%Z"] = tz_datetime.strftime("%Z")
        else:
            replacement_dict["%z"] = ""
            replacement_dict["%Z"] = ""

        # Hack together a "Locale's appropriate date and time representation."
        replacement_dict["%c"] = self._native_date.strftime("%c").replace(
            "00:00:00", native_time.strftime("%X")
        )

        # Build the picosecond, attosecond, and yoctosecond replacements
        nanosecond_string = util.decimal_stringify(self._attotime._nanosecond)

        if "." in nanosecond_string:
            stop_index = nanosecond_string.index(".")

            left_padding = 3 - stop_index
            right_padding = 18 - len(nanosecond_string[stop_index:])
        else:
            left_padding = 3 - len(nanosecond_string)
            right_padding = 18 - left_padding - len(nanosecond_string)

        nanosecond_string = "0" * left_padding + nanosecond_string + "0" * right_padding
        nanosecond_string = nanosecond_string.replace(".", "")

        replacement_dict["%o"] = nanosecond_string[0:6]  # picoseconds
        replacement_dict["%q"] = nanosecond_string[6:12]  # attoseconds
        replacement_dict["%v"] = nanosecond_string[12:18]  # yoctoseconds

        replacement_dict["%%"] = "%"

        return util.multiple_replace(formatstr, replacement_dict)

    def __add__(self, y):
        # Return the result of x + y, where self is x and y is an attotimedelta
        if not isinstance(y, attotimedelta):
            return NotImplemented

        delta = attotimedelta(
            days=self._native_date.toordinal(),
            hours=self.hour,
            minutes=self.minute,
            seconds=self.second,
            microseconds=self.microsecond,
            nanoseconds=self.nanosecond,
        )

        delta += y

        hour, remaining_seconds = getcontext().divmod(
            delta.seconds, constants.SECONDS_PER_HOUR
        )
        minute, second = getcontext().divmod(
            remaining_seconds, constants.SECONDS_PER_MINUTE
        )

        if delta.days > 0 and delta.days < datetime.date.max.toordinal():
            return self.combine(
                datetime.date.fromordinal(delta.days),
                attotime(
                    int(hour),
                    int(minute),
                    int(second),
                    delta.microseconds,
                    delta.nanoseconds,
                    tzinfo=self.tzinfo,
                ),
            )

        raise OverflowError("result out of range")

    def __sub__(self, y):
        # Returns the result of x - y, where self ix x, and y is
        # an attodatetime or attotimedelta, when y is an attodatetime,
        # the result is an attotimedelta, when y is an attotimedelta,
        # the result is an attodatetime
        if not isinstance(y, attodatetime):
            if isinstance(y, attotimedelta):
                # y is an attotimedelta
                return self + -y

            return NotImplemented

        # y is an attodatetime, build an attotimedelta
        daysx = self._native_date.toordinal()
        daysy = y._native_date.toordinal()

        secondsx = (
            self.second
            + self.minute * constants.SECONDS_PER_MINUTE
            + self.hour * constants.SECONDS_PER_HOUR
        )
        secondsy = (
            y.second
            + y.minute * constants.SECONDS_PER_MINUTE
            + y.hour * constants.SECONDS_PER_HOUR
        )

        microsecondsx = self.microsecond
        microsecondsy = y.microsecond

        nanosecondsx = self.nanosecond
        nanosecondsy = y.nanosecond

        base = attotimedelta(
            days=daysx - daysy,
            seconds=secondsx - secondsy,
            microseconds=microsecondsx - microsecondsy,
            nanoseconds=nanosecondsx - nanosecondsy,
        )

        if self.tzinfo is y.tzinfo:
            return base

        offsetx = self.utcoffset()
        offsety = y.utcoffset()

        if offsetx == offsety:
            return base

        if offsetx is None or offsety is None:
            raise TypeError("cannot mix naive and timezone-aware time")

        return base + offsety - offsetx

    def __eq__(self, y):
        # Return the result of x == y, where self is x
        if isinstance(y, self.__class__):
            return self._native_date == y._native_date and self._attotime == y._attotime

        return False

    def __ne__(self, y):
        # Return the result of x != y, where self is x
        return not self.__eq__(y)

    def __gt__(self, y):
        # Return the result of x > y, where self is x
        if isinstance(y, self.__class__):
            return self._native_date > y._native_date or (
                self._native_date == y._native_date and self._attotime > y._attotime
            )
        else:
            raise TypeError(
                "can't compare {0} to {1}".format(type(self).__name__, type(y).__name__)
            )

    def __ge__(self, y):
        # Return the result of x >= y, where self is x
        if isinstance(y, self.__class__):
            return self._native_date > y._native_date or (
                self._native_date == y._native_date and self._attotime >= y._attotime
            )
        else:
            raise TypeError(
                "can't compare {0} to {1}".format(type(self).__name__, type(y).__name__)
            )

    def __lt__(self, y):
        # Return the result of x < y, where self is x
        if isinstance(y, self.__class__):
            return self._native_date < y._native_date or (
                self._native_date == y._native_date and self._attotime < y._attotime
            )
        else:
            raise TypeError(
                "can't compare {0} to {1}".format(type(self).__name__, type(y).__name__)
            )

    def __le__(self, y):
        # Return the result of x <= y, where self is x
        if isinstance(y, self.__class__):
            return self._native_date < y._native_date or (
                self._native_date == y._native_date and self._attotime <= y._attotime
            )
        else:
            raise TypeError(
                "can't compare {0} to {1}".format(type(self).__name__, type(y).__name__)
            )

    def __str__(self):
        return self.isoformat(separator=" ")

    def __repr__(self):
        if self._attotime.tzinfo is None:
            return "{0}({1}, {2}, {3}, {4}, {5}, {6}, {7}, {8})".format(
                self.__class__.__module__,
                self._native_date.year,
                self._native_date.month,
                self._native_date.day,
                self._attotime.hour,
                self._attotime.minute,
                self._attotime.second,
                self._attotime.microsecond,
                str(self._attotime.nanosecond),
            )

        return "{0}({1}, {2}, {3}, {4}, {5}, {6}, {7}, {8}, {9})".format(
            self.__class__.__module__,
            self._native_date.year,
            self._native_date.month,
            self._native_date.day,
            self._attotime.hour,
            self._attotime.minute,
            self._attotime.second,
            self._attotime.microsecond,
            str(self._attotime.nanosecond),
            str(self._attotime.tzinfo),
        )

    def __format__(self, formatstr):
        return self.strftime(formatstr)

    def _as_datetime(self):
        # Returns the attodatetime as a native datetime, losing nanosecond resolution
        return datetime.datetime(
            year=self.year,
            month=self.month,
            day=self.day,
            hour=self.hour,
            minute=self.minute,
            second=self.second,
            microsecond=self.microsecond,
            tzinfo=self.tzinfo,
        )



# -*- coding: utf-8 -*-

# Copyright (c) 2021, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

import calendar
import time

from attotime import compat


def get_format_fields(format):
    # Return a list of format fields
    component_accumulator = ""
    format_components = []

    for field_char in format:
        if component_accumulator.startswith("%") and len(component_accumulator) == 2:
            # Field complete, reset accumulator
            format_components.append(component_accumulator)
            component_accumulator = ""
        elif (
            field_char == "%"
            and len(component_accumulator) > 0
            and not component_accumulator.startswith("%")
        ):
            # Starting new field, reset accumulator
            format_components.append(component_accumulator)
            component_accumulator = ""

        component_accumulator += field_char

    if component_accumulator != "":
        # Grab whatever was in the accumulator
        format_components.append(component_accumulator)

    return format_components


def expand_format_fields(format_fields):
    # Expands %c, %x, %X into their component fields
    # Taken from the Python implementation
    # https://github.com/python/cpython/blob/master/Lib/_strptime.py#L107
    expanded_format_fields = []

    # First we need timezone strings
    # https://github.com/python/cpython/blob/master/Lib/_strptime.py#L152
    tzname = time.tzname
    daylight = time.daylight
    no_saving = frozenset({"utc", "gmt", tzname[0].lower()})

    if daylight:
        has_saving = frozenset({tzname[1].lower()})
    else:
        has_saving = frozenset()

    timezone_tuple = (no_saving, has_saving)

    # Day and month strings
    a_string = calendar.day_abbr[2].lower()
    A_string = calendar.day_name[2].lower()

    b_string = calendar.month_abbr[3].lower()
    B_string = calendar.month_name[3].lower()

    # AM/PM string
    pm_time_tuple = time.struct_time((1999, 3, 17, 22, 44, 55, 2, 76, 0))
    pm_string = time.strftime("%p", pm_time_tuple).lower()

    # Now calculate the expansions
    time_tuple = time.struct_time((1999, 3, 17, 22, 44, 55, 2, 76, 0))

    expansions = [None, None, None]
    expansions[0] = time.strftime("%c", time_tuple).lower()
    expansions[1] = time.strftime("%x", time_tuple).lower()
    expansions[2] = time.strftime("%X", time_tuple).lower()

    replacement_pairs = [
        ("%", "%%"),
        (A_string, "%A"),
        (B_string, "%B"),
        (a_string, "%a"),
        (b_string, "%b"),
        (pm_string, "%p"),
        ("1999", "%Y"),
        ("99", "%y"),
        ("22", "%H"),
        ("44", "%M"),
        ("55", "%S"),
        ("76", "%j"),
        ("17", "%d"),
        ("03", "%m"),
        ("3", "%m"),
        ("2", "%w"),
        ("10", "%I"),
    ]

    replacement_pairs.extend(
        [(tz, "%Z") for tz_values in timezone_tuple for tz in tz_values]
    )

    for offset, directive in ((0, "%c"), (1, "%x"), (2, "%X")):
        current_format = expansions[offset]

        for old, new in replacement_pairs:
            current_format = current_format.replace(old, new)

        time_tuple = time.struct_time((1999, 1, 3, 1, 1, 1, 6, 3, 0))

        if "00" in time.strftime(directive, time_tuple):
            U_W = "%W"
        else:
            U_W = "%U"

        expansions[offset] = current_format.replace("11", U_W)

    c_expansion = get_format_fields(expansions[0])
    x_expansion = get_format_fields(expansions[1])
    X_expansion = get_format_fields(expansions[2])

    # Finally, perform the expansion
    for format_field in format_fields:
        if format_field == "%c":
            expanded_format_fields += c_expansion
        elif format_field == "%x":
            expanded_format_fields += x_expansion
        elif format_field == "%X":
            expanded_format_fields += X_expansion
        else:
            expanded_format_fields.append(format_field)

    return expanded_format_fields


def get_field_size(date_string, field_members, range_start, range_end):
    # Given the date string, a list of field members, and the range to start
    # in, find the member in the range, and return its size
    max_field_size = 0

    for field in field_members:
        if len(field) > max_field_size:
            max_field_size = len(field)

    for range_index in compat.range(range_start, range_end + max_field_size):
        for field_member in field_members:
            if date_string.lower()[range_index:].startswith(field_member.lower()):
                return len(field_member)

    # Field not found
    return -1



# -*- coding: utf-8 -*-

# Copyright (c) 2021, Brandon Nielsen
# All rights reserved.
#
# This software may be modified and distributed under the terms
# of the BSD license.  See the LICENSE file for details.

import time
import unittest

from attotime.tests.compat import mock
from attotime.util import strptime

# Used for forcing a locale
LOCALE_DAY_ABBRS = ["Mon", "Tue", "Wed", "Thu", "Fri", "Sat", "Sun"]
LOCALE_DAY_NAMES = [
    "Monday",
    "Tuesday",
    "Wednesday",
    "Thursday",
    "Friday",
    "Saturday",
    "Sunday",
]

LOCALE_MONTH_ABBRS = [
    "",
    "Jan",
    "Feb",
    "Mar",
    "Apr",
    "May",
    "Jun",
    "Jul",
    "Aug",
    "Sep",
    "Oct",
    "Nov",
    "Dec",
]
LOCALE_MONTH_NAMES = [
    "",
    "January",
    "February",
    "March",
    "April",
    "May",
    "June",
    "July",
    "August",
    "September",
    "October",
    "November",
    "December",
]


def _mock_strftime(format, timetuple):
    if timetuple == time.struct_time((1999, 3, 17, 22, 44, 55, 2, 76, 0)):
        if format == "%p":
            return "PM"
        elif format == "%c":
            return "Wed Mar 17 22:44:55 1999"
        elif format == "%x":
            return "03/17/99"
        elif format == "%X":
            return "22:44:55"
    elif timetuple == time.struct_time((1999, 1, 3, 1, 1, 1, 6, 3, 0)):
        if format == "%c":
            return "Sun Jan 3 01:01:01 1999"
        elif format == "%x":
            return "01/03/99"
        elif format == "%X":
            return "01:01:01"


class TestStrptimeFunctions(unittest.TestCase):
    def test_get_format_fields(self):
        self.assertEqual(
            strptime.get_format_fields(
                "%a%A%w%d%b%B%m%y%Y%H%I%p%M%S%f%z%Z%j%U%W%c%x%X%%"
            ),
            [
                "%a",
                "%A",
                "%w",
                "%d",
                "%b",
                "%B",
                "%m",
                "%y",
                "%Y",
                "%H",
                "%I",
                "%p",
                "%M",
                "%S",
                "%f",
                "%z",
                "%Z",
                "%j",
                "%U",
                "%W",
                "%c",
                "%x",
                "%X",
                "%%",
            ],
        )
        self.assertEqual(strptime.get_format_fields("%o%q%v"), ["%o", "%q", "%v"])
        self.assertEqual(strptime.get_format_fields("abcd"), ["abcd"])
        self.assertEqual(
            strptime.get_format_fields("%a abcd %X"), ["%a", " abcd ", "%X"]
        )

        self.assertEqual(strptime.get_format_fields(""), [])

    @mock.patch(
        "attotime.util.strptime.calendar.day_abbr",
        new_callable=mock.PropertyMock(return_value=LOCALE_DAY_ABBRS),
    )
    @mock.patch(
        "attotime.util.strptime.calendar.day_name",
        new_callable=mock.PropertyMock(return_value=LOCALE_DAY_NAMES),
    )
    @mock.patch(
        "attotime.util.strptime.calendar.month_abbr",
        new_callable=mock.PropertyMock(return_value=LOCALE_MONTH_ABBRS),
    )
    @mock.patch(
        "attotime.util.strptime.calendar.month_name",
        new_callable=mock.PropertyMock(return_value=LOCALE_MONTH_NAMES),
    )
    @mock.patch("attotime.util.strptime.time.strftime")
    def test_expand_format_fields(
        self, mockStrftime, mockMonthName, mockMonthAbbr, mockDayName, mockDayAbbr
    ):
        self.assertEqual(
            strptime.expand_format_fields(
                [
                    "%a",
                    "%A",
                    "%w",
                    "%d",
                    "%b",
                    "%B",
                    "%m",
                    "%y",
                    "%Y",
                    "%H",
                    "%I",
                    "%p",
                    "%M",
                    "%S",
                    "%f",
                    "%z",
                    "%Z",
                    "%j",
                    "%U",
                    "%W",
                    "%%",
                ]
            ),
            [
                "%a",
                "%A",
                "%w",
                "%d",
                "%b",
                "%B",
                "%m",
                "%y",
                "%Y",
                "%H",
                "%I",
                "%p",
                "%M",
                "%S",
                "%f",
                "%z",
                "%Z",
                "%j",
                "%U",
                "%W",
                "%%",
            ],
        )
        self.assertEqual(
            strptime.expand_format_fields(["%o", "%q", "%v"]), ["%o", "%q", "%v"]
        )

        mockStrftime.side_effect = _mock_strftime

        self.assertEqual(
            strptime.expand_format_fields(["%c"]),
            [
                "%a",
                " ",
                "%b",
                " ",
                "%d",
                " ",
                "%H",
                ":",
                "%M",
                ":",
                "%S",
                " ",
                "%Y",
            ],
        )
        self.assertEqual(
            strptime.expand_format_fields(["%x"]),
            ["%m", "/", "%d", "/", "%y"],
        )
        self.assertEqual(
            strptime.expand_format_fields(["%X"]),
            ["%H", ":", "%M", ":", "%S"],
        )
        self.assertEqual(
            strptime.expand_format_fields(["%c", " ", "%o", "%q", "%v"]),
            [
                "%a",
                " ",
                "%b",
                " ",
                "%d",
                " ",
                "%H",
                ":",
                "%M",
                ":",
                "%S",
                " ",
                "%Y",
                " ",
                "%o",
                "%q",
                "%v",
            ],
        )

    @mock.patch(
        "attotime.util.strptime.calendar.day_abbr",
        new_callable=mock.PropertyMock(return_value=LOCALE_DAY_ABBRS),
    )
    @mock.patch(
        "attotime.util.strptime.calendar.day_name",
        new_callable=mock.PropertyMock(return_value=LOCALE_DAY_NAMES),
    )
    @mock.patch(
        "attotime.util.strptime.calendar.month_abbr",
        new_callable=mock.PropertyMock(return_value=LOCALE_MONTH_ABBRS),
    )
    @mock.patch(
        "attotime.util.strptime.calendar.month_name",
        new_callable=mock.PropertyMock(return_value=LOCALE_MONTH_NAMES),
    )
    @mock.patch(
        "attotime.util.strptime.time.daylight",
        new_callable=mock.PropertyMock(return_value=False),
    )
    @mock.patch("attotime.util.strptime.time.strftime")
    def test_expand_format_fields_daylight(
        self,
        mockStrftime,
        mockDaylight,
        mockMonthName,
        mockMonthAbbr,
        mockDayName,
        mockDayAbbr,
    ):
        mockStrftime.side_effect = _mock_strftime

        self.assertEqual(
            strptime.expand_format_fields(["%c"]),
            [
                "%a",
                " ",
                "%b",
                " ",
                "%d",
                " ",
                "%H",
                ":",
                "%M",
                ":",
                "%S",
                " ",
                "%Y",
            ],
        )
        self.assertEqual(
            strptime.expand_format_fields(["%x"]),
            ["%m", "/", "%d", "/", "%y"],
        )
        self.assertEqual(
            strptime.expand_format_fields(["%X"]),
            ["%H", ":", "%M", ":", "%S"],
        )
        self.assertEqual(
            strptime.expand_format_fields(["%c", " ", "%o", "%q", "%v"]),
            [
                "%a",
                " ",
                "%b",
                " ",
                "%d",
                " ",
                "%H",
                ":",
                "%M",
                ":",
                "%S",
                " ",
                "%Y",
                " ",
                "%o",
                "%q",
                "%v",
            ],
        )

    @mock.patch(
        "attotime.util.strptime.calendar.day_abbr",
        new_callable=mock.PropertyMock(return_value=LOCALE_DAY_ABBRS),
    )
    @mock.patch(
        "attotime.util.strptime.calendar.day_name",
        new_callable=mock.PropertyMock(return_value=LOCALE_DAY_NAMES),
    )
    @mock.patch(
        "attotime.util.strptime.calendar.month_abbr",
        new_callable=mock.PropertyMock(return_value=LOCALE_MONTH_ABBRS),
    )
    @mock.patch(
        "attotime.util.strptime.calendar.month_name",
        new_callable=mock.PropertyMock(return_value=LOCALE_MONTH_NAMES),
    )
    @mock.patch("attotime.util.strptime.time.strftime")
    def test_expand_format_fields_weeknumber(
        self, mockStrftime, mockMonthName, mockMonthAbbr, mockDayName, mockDayAbbr
    ):
        def _mock_strftime_weeknumber(format, timetuple):
            if timetuple == time.struct_time((1999, 3, 17, 22, 44, 55, 2, 76, 0)):
                if format == "%p":
                    return "PM"
                elif format == "%c":
                    return "Wed Mar 17 22:44:55 1999"
                elif format == "%x":
                    return "03/17/99"
                elif format == "%X":
                    return "22:44:55"
            elif timetuple == time.struct_time((1999, 1, 3, 1, 1, 1, 6, 3, 0)):
                if format == "%c":
                    return "00 0 01:01:01 1999"
                elif format == "%x":
                    return "01/03/99"
                elif format == "%X":
                    return "01:01:01"

        mockStrftime.side_effect = _mock_strftime_weeknumber

        self.assertEqual(
            strptime.expand_format_fields(["%c"]),
            [
                "%a",
                " ",
                "%b",
                " ",
                "%d",
                " ",
                "%H",
                ":",
                "%M",
                ":",
                "%S",
                " ",
                "%Y",
            ],
        )
        self.assertEqual(
            strptime.expand_format_fields(["%x"]),
            ["%m", "/", "%d", "/", "%y"],
        )
        self.assertEqual(
            strptime.expand_format_fields(["%X"]),
            ["%H", ":", "%M", ":", "%S"],
        )
        self.assertEqual(
            strptime.expand_format_fields(["%c", " ", "%o", "%q", "%v"]),
            [
                "%a",
                " ",
                "%b",
                " ",
                "%d",
                " ",
                "%H",
                ":",
                "%M",
                ":",
                "%S",
                " ",
                "%Y",
                " ",
                "%o",
                "%q",
                "%v",
            ],
        )

    def test_get_field_size(self):
        self.assertEqual(strptime.get_field_size("2001", ["2001"], 0, 1), 4)
        self.assertEqual(strptime.get_field_size("abcd2001", ["2001"], 0, 7), 4)
        self.assertEqual(strptime.get_field_size("abcd2001", ["2001"], 0, 5), 4)

        self.assertEqual(strptime.get_field_size("abcd2001", ["BC"], 0, 7), 2)
        self.assertEqual(strptime.get_field_size("abcd2001", ["BC"], 0, 2), 2)

        self.assertEqual(
            strptime.get_field_size("2001", ["a", "b", "c", "d"], 0, 2), -1
        )
        self.assertEqual(strptime.get_field_size("2001", ["2001"], 1, 2), -1)
